from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, Callable
from unittest import mock

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loader import open as loader_open
from dissect.target.loaders.hyperv import HyperVLoader
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``HyperVLoader`` when opening a ``Target``."""
    for path in [
        absolute_path("_data/loaders/hyperv/B90AC31B-C6F8-479F-9B91-07B894A6A3F6.xml"),
        absolute_path("_data/loaders/hyperv/D351C151-DAC7-4042-B434-B72D522C1E4A.xml"),
        absolute_path("_data/loaders/hyperv/EC04F346-DB96-4700-AF5B-77B3C56C38BD.vmcx"),
        absolute_path("_data/loaders/hyperv/993F7B33-6057-4D1E-A1FE-A1A1D77BE974.vmcx"),
    ]:
        with mock.patch("dissect.target.target.Target.apply"):
            target = opener(path)
            assert isinstance(target._loader, HyperVLoader)
            assert target.path == path


@pytest.mark.parametrize(
    ("descriptor_dir", "disk_dir"),
    [
        # Disk next to descriptor file
        ("c:", "c:"),
        # Disk in "Virtual Hard Disks" directory, relative from descriptor file
        ("c:", "c:\\Virtual Hard Disks"),
        # Disk in "Virtual Hard Disks" directory, one directory up relative from descriptor file
        ("c:\\Virtual Machines", "c:\\Virtual Hard Disks"),
        # Disk in absolute location
        ("c:\\Virtual Machines", "c:\\Disks"),
    ],
)
def test_loader(descriptor_dir: str, disk_dir: str, target_bare: Target) -> None:
    """Test the Hyper-V loader with XML and VMCX descriptor files."""
    gen1_xml_filename = "B90AC31B-C6F8-479F-9B91-07B894A6A3F6.xml"
    gen2_xml_filename = "D351C151-DAC7-4042-B434-B72D522C1E4A.xml"
    gen1_vmcx_filename = "EC04F346-DB96-4700-AF5B-77B3C56C38BD.vmcx"
    gen2_vmcx_filename = "993F7B33-6057-4D1E-A1FE-A1A1D77BE974.vmcx"

    vfs = VirtualFilesystem(case_sensitive=False, alt_separator="\\")
    vfs.map_file(f"{descriptor_dir}\\{gen1_xml_filename}", absolute_path(f"_data/loaders/hyperv/{gen1_xml_filename}"))
    vfs.map_file(f"{descriptor_dir}\\{gen2_xml_filename}", absolute_path(f"_data/loaders/hyperv/{gen2_xml_filename}"))
    vfs.map_file(f"{descriptor_dir}\\{gen1_vmcx_filename}", absolute_path(f"_data/loaders/hyperv/{gen1_vmcx_filename}"))
    vfs.map_file(f"{descriptor_dir}\\{gen2_vmcx_filename}", absolute_path(f"_data/loaders/hyperv/{gen2_vmcx_filename}"))

    # Fake disks
    vfs.map_file_fh(f"{disk_dir}\\Default Generation 1.vhdx", BytesIO())
    vfs.map_file_fh(f"{disk_dir}\\Default Generation 2.vhdx", BytesIO())

    gen1_xml_path = vfs.path(f"{descriptor_dir}\\{gen1_xml_filename}")
    gen2_xml_path = vfs.path(f"{descriptor_dir}\\{gen2_xml_filename}")
    gen1_vmcx_path = vfs.path(f"{descriptor_dir}\\{gen1_vmcx_filename}")
    gen2_vmcx_path = vfs.path(f"{descriptor_dir}\\{gen2_vmcx_filename}")

    with mock.patch("dissect.target.container.open") as mock_container_open:
        loader = loader_open(gen1_xml_path)
        assert isinstance(loader, HyperVLoader)

        loader.map(target_bare)
        mock_container_open.assert_called_with(vfs.path(f"{disk_dir}\\Default Generation 1.vhdx").resolve())

    with mock.patch("dissect.target.container.open") as mock_container_open:
        loader = loader_open(gen2_xml_path)
        assert isinstance(loader, HyperVLoader)

        loader.map(target_bare)
        mock_container_open.assert_called_with(vfs.path(f"{disk_dir}\\Default Generation 2.vhdx").resolve())

    with mock.patch("dissect.target.container.open") as mock_container_open:
        loader = loader_open(gen1_vmcx_path)
        assert isinstance(loader, HyperVLoader)

        loader.map(target_bare)
        mock_container_open.assert_called_with(vfs.path(f"{disk_dir}\\Default Generation 1.vhdx").resolve())

    with mock.patch("dissect.target.container.open") as mock_container_open:
        loader = loader_open(gen2_vmcx_path)
        assert isinstance(loader, HyperVLoader)

        loader.map(target_bare)
        mock_container_open.assert_called_with(vfs.path(f"{disk_dir}\\Default Generation 2.vhdx").resolve())



from __future__ import annotations

import platform

import pytest

from dissect.target.filesystems.squashfs import SquashFSFilesystem
from tests._utils import absolute_path


@pytest.mark.parametrize("type", ["gzip", "gzip-opts", "lz4", "lzma", "lzo", "xz", "zstd"])
def test_filesystem_squashfs(type: str) -> None:
    """Test if we detect and correctly iterate SquashFS filesystems."""

    if type == "lzma" and platform.system() == "Windows":
        pytest.skip(reason="SquashFS LZMA misbehaves on Windows")

    sqfs_bin = absolute_path(f"_data/filesystems/squashfs/{type}.sqfs")

    with sqfs_bin.open("rb") as fh:
        assert SquashFSFilesystem.detect(fh)

        fs = SquashFSFilesystem(fh)

        assert [f.name for f in fs.path("/").iterdir()] == [
            "dir",
            "file-with-xattr",
            "large-file",
            "small-file",
            "symlink-1",
            "symlink-2",
            "symlink-with-xattr",
        ]

        assert fs.path("/dir").is_dir()
        assert fs.path("/symlink-2").is_symlink()
        assert fs.path("/dir/file_99").is_file()
        assert fs.path("/small-file").read_bytes() == b"contents\n"
        assert fs.path("/symlink-2").resolve().as_posix() == "/dir/file_69"

        stat = fs.path("/large-file").lstat()
        assert stat.st_mode == 0o100644
        assert stat.st_ino == 103
        assert stat.st_nlink == 1
        assert stat.st_uid == 1000
        assert stat.st_gid == 1000
        assert stat.st_size == 4177920
        assert stat.st_mtime == 1670266447.0



from __future__ import annotations

import io
import logging
import platform
import urllib.parse
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Callable, ClassVar
from unittest.mock import MagicMock, Mock, PropertyMock, call, patch

import pytest

from dissect.target import loader
from dissect.target.containers.raw import RawContainer
from dissect.target.exceptions import FilesystemError
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.loaders.dir import DirLoader
from dissect.target.loaders.raw import RawLoader
from dissect.target.loaders.vbox import VBoxLoader
from dissect.target.target import DiskCollection, Event, Target, TargetLogAdapter, log

if TYPE_CHECKING:
    from collections.abc import Iterator
    from logging import Logger


class ErrorCounter(TargetLogAdapter):
    errors = 0

    def error(*args, **kwargs) -> None:
        ErrorCounter.errors += 1


@pytest.mark.parametrize(
    ("topo", "entry_point", "expected_result", "expected_errors"),
    [
        # Base cases:
        # Scenario 0:  Can we still load a single target that's not a dir?
        (
            [
                "/raw.tst",
            ],
            "/raw.tst",
            "[TestLoader('/raw.tst')]",
            0,
        ),
        # Scenario 0b: Make sure we still get so see errors if a target fails to load
        # (1 error + 1 exception no load)
        (
            [
                "/raw.vbox",
            ],
            "/raw.vbox",
            "[]",
            2,
        ),
        # Scenario 0c: Attempting to load a dir with nothing of interest yield an error
        # (1 exception no load)
        (
            [
                "/dir/garbage",
            ],
            "/dir",
            "[]",
            1,
        ),
        # Dir cases:
        # Scenario 1:  Dir is loadable target (simple)
        (
            [
                "/dir/etc",
                "/dir/var",
            ],
            "/dir",
            "[DirLoader('/dir')]",
            0,
        ),
        # Scenario 2: dir contains multiple loadable targets
        (
            [
                "/dir/raw.img",
                "/dir/raw2.tst",
            ],
            "/dir",
            "[RawLoader('/dir/raw.img'), TestLoader('/dir/raw2.tst')]",
            0,
        ),
        # Scenario 2b: dir contains multiple loadable targets and some garbage
        # (it will wrap the garbage in a RawLoader to try)
        (
            [
                "/dir/raw.img",
                "/dir/raw2.tst",
                "/dir/info.txt",
            ],
            "/dir",
            "[RawLoader('/dir/raw.img'), TestLoader('/dir/raw2.tst'), RawLoader('/dir/info.txt')]",
            0,
        ),
        # Scenario 3: dir contains 1 loadable dir as well as 2 loadable targets
        (
            [
                "/dir/unix/etc",
                "/dir/unix/var",
                "/dir/raw.img",
                "/dir/raw2.tst",
            ],
            "/dir",
            "[DirLoader('/dir/unix'), RawLoader('/dir/raw.img'), TestLoader('/dir/raw2.tst')]",
            0,
        ),
        # Scenario 3b: dir contains 2 loadable dirs as well as 2 loadable targets
        (
            [
                "/dir/unix/etc",
                "/dir/unix/var",
                "/dir/win/c:",
                "/dir/win/c:/windows/system32",
                "/dir/raw.img",
                "/dir/raw2.tst",
            ],
            "/dir",
            "[DirLoader('/dir/unix'), DirLoader('/dir/win'), RawLoader('/dir/raw.img'), TestLoader('/dir/raw2.tst')]",
            0,
        ),
        # Scenario 4: dir is a loadable dir but contains other files including loadable targets
        (
            [
                "/dir/etc",
                "/dir/var",
                "/dir/raw.img",
                "/dir/info.txt",
            ],
            "/dir",
            "[DirLoader('/dir')]",
            0,
        ),
        # Scenario 4b: Windows variant
        (
            [
                "/dir/c:",
                "/dir/c:/windows/system32",
                "/dir/raw.img",
                "/dir/info.txt",
            ],
            "/dir",
            "[DirLoader('/dir')]",
            0,
        ),
        # Scenario 5: Hypothetical Dirloader with selection
        (
            [
                "/dir/select.txt",  # selects 1 and 3
                "/dir/raw1.img",
                "/dir/raw2.img",
                "/dir/raw3.img",
            ],
            "/dir",
            "[SelectLdr('/dir/raw1.img'), SelectLdr('/dir/raw3.img')]",
            0,
        ),
    ],
)
@patch("dissect.target.target.getlogger", new=lambda t: ErrorCounter(log, {"target": t}))
def test_target_open_dirs(topo: list[str], entry_point: str, expected_result: str, expected_errors: int) -> None:
    # TestLoader to mix Raw Targets with Test Targets without depending on
    # specific implementations.
    class TestLoader(RawLoader):
        @staticmethod
        def detect(path: Path) -> bool:
            return str(path).endswith(".tst")

        def map(self, target: Target) -> None:
            target.disks.add(RawContainer(io.BytesIO(b"\x00")))

    class SelectLdr(DirLoader):
        @staticmethod
        def detect(path: Path) -> bool:
            return path.is_dir() and path.joinpath("select.txt").exists()

        @staticmethod
        def find_all(path: Path, **kwargs) -> Iterator[Path]:
            return [path.joinpath("/dir/raw1.img"), path.joinpath("/dir/raw3.img")]

        def map(self, target: Target) -> None:
            target.disks.add(RawContainer(io.BytesIO(b"\x00")))

    def make_vfs(topo: dict) -> dict:
        vfs = VirtualFilesystem()
        for entry in topo:
            if entry.find(".") > -1:
                vfs.map_file_fh(entry, io.BytesIO(b"\x00"))
            else:
                vfs.makedirs(entry)
        return vfs

    def dirtest(vfs: dict, selector: str) -> list:
        ErrorCounter.errors = 0
        try:
            return [x._loader for x in Target.open_all([vfs.path(selector)])]
        except Exception:
            ErrorCounter.errors += 1
            return []

    with patch.object(loader, "LOADERS", [VBoxLoader, TestLoader, SelectLdr]):
        assert str(dirtest(make_vfs(topo), entry_point)) == expected_result

        assert ErrorCounter.errors == expected_errors


@pytest.fixture
def mocked_win_volumes_fs() -> Iterator[tuple[Mock, Mock, Mock]]:
    mock_bad_volume = Mock(name="bad-volume-Z")
    mock_bad_volume.fs = None
    mock_bad_volume.drive_letter = "Z"

    mock_good_volume = Mock(name="good-volume-W")
    mock_good_volume.fs = None
    mock_good_volume.drive_letter = "W"

    mock_good_fs = Mock(name="good-fs")
    mock_good_fs.iter_subfs.return_value = []

    def mock_filesystem_open(volume: Mock) -> Mock:
        if volume == mock_good_volume:
            return mock_good_fs
        raise FilesystemError("Not supported")

    def get_mock_drive_volumes(log: Logger) -> list[tuple[Mock, bool, Mock, int]]:
        # drive_volume, is_windrive, disk, disk_num
        return [
            (mock_good_volume, False, Mock(), 1),
            (mock_bad_volume, False, Mock(), 2),
        ]

    def mock_disk_apply(self: DiskCollection) -> None:
        if len(self.target.disks) > 0:
            self.target.volumes.add(mock_good_volume)
            self.target.volumes.add(mock_bad_volume)

    with (
        patch("dissect.target.loaders.local._get_os_name") as mock_os_name,
        patch("dissect.target.filesystem.open", new=mock_filesystem_open),
        patch("dissect.target.plugin.os_plugins") as mock_os_plugins,
        patch("dissect.target.target.DiskCollection.apply", new=mock_disk_apply),
        patch("dissect.target.loaders.local._get_windows_drive_volumes", new=get_mock_drive_volumes),
    ):
        mock_os_plugins.return_value = []
        mock_os_name.return_value = "windows"

        yield (mock_good_volume, mock_bad_volume, mock_good_fs)


@pytest.fixture
def mocked_lin_volumes_fs() -> Iterator[tuple[Mock, Mock, Mock]]:
    mock_bad_volume = Mock(name="bad-volume")
    mock_bad_volume.fs = None

    mock_good_volume = Mock(name="good-volume")
    mock_good_volume.fs = None

    mock_good_fs = Mock(name="good-fs")

    def mock_filesystem_open(volume: Mock) -> Mock:
        if volume == mock_good_volume:
            return mock_good_fs
        raise FilesystemError("Not supported")

    def mock_disk_apply(self: DiskCollection) -> None:
        if len(self.target.disks) > 0:
            self.target.volumes.add(mock_good_volume)
            self.target.volumes.add(mock_bad_volume)

    with (
        patch("dissect.target.loaders.local._get_os_name") as mock_os_name,
        patch("dissect.target.filesystem.open", new=mock_filesystem_open),
        patch("dissect.target.plugin.os_plugins") as mock_os_plugins,
        patch("dissect.target.target.DiskCollection.apply", new=mock_disk_apply),
    ):
        mock_os_plugins.return_value = []
        mock_os_name.return_value = "linux"

        yield (mock_good_volume, mock_bad_volume, mock_good_fs)


def test_target_win_force_dirfs(mocked_win_volumes_fs: tuple[Mock, Mock, Mock]) -> None:
    mock_good_volume, mock_bad_volume, mock_good_fs = mocked_win_volumes_fs

    query = {"force-directory-fs": 1}
    target_query = urllib.parse.urlencode(query)
    force_dirfs_path = f"local?{target_query}"
    target = Target.open(force_dirfs_path)

    assert "force-directory-fs" in target.path_query

    # original volumes are attached
    assert len(target.volumes) == 2

    # FSs from both volumes, bad and good, are mounted + 2 mounts by DefaultOSPlugin
    assert set(target.fs.mounts.keys()) == {
        mock_bad_volume.drive_letter,
        mock_good_volume.drive_letter,
        "fs0",
        "fs1",
    }

    # both volumes should be mounted as DirectoryFilesystem
    assert isinstance(target.fs.mounts[mock_good_volume.drive_letter], DirectoryFilesystem)
    assert isinstance(target.fs.mounts[mock_bad_volume.drive_letter], DirectoryFilesystem)


def test_target_win_fallback_dirfs(mocked_win_volumes_fs: tuple[Mock, Mock, Mock]) -> None:
    mock_good_volume, mock_bad_volume, mock_good_fs = mocked_win_volumes_fs

    query = {"fallback-to-directory-fs": 1}
    target_query = urllib.parse.urlencode(query)
    fallback_dirfs_path = f"local?{target_query}"
    target = Target.open(fallback_dirfs_path)

    assert "fallback-to-directory-fs" in target.path_query

    # both volumes are present
    assert len(target.volumes) == 2

    # FSs from both volumes, bad and good, are mounted + 2 mounts by DefaultOSPlugin
    assert set(target.fs.mounts.keys()) == {
        mock_bad_volume.drive_letter,
        mock_good_volume.drive_letter,
        "fs0",
        "fs1",
    }

    # One volume is mounted as DirectoryFilesystem, another as original mock_good_fs
    assert target.fs.mounts[mock_good_volume.drive_letter] == mock_good_fs
    assert isinstance(target.fs.mounts[mock_bad_volume.drive_letter], DirectoryFilesystem)


def test_target_force_dirfs_linux(mocked_lin_volumes_fs: tuple[Mock, Mock, Mock]) -> None:
    query = {"force-directory-fs": 1}
    target_query = urllib.parse.urlencode(query)
    force_dirfs_path = f"local?{target_query}"
    target = Target.open(force_dirfs_path)

    assert "force-directory-fs" in target.path_query

    # no original volumes should be attached
    assert len(target.volumes) == 0

    assert len(target.fs.mounts) == 1
    assert target.fs.mounts.get("/")
    assert isinstance(target.fs.mounts["/"], DirectoryFilesystem)


def test_target_fallback_dirfs_linux(mocked_lin_volumes_fs: tuple[Mock, Mock, Mock]) -> None:
    query = {"fallback-to-directory-fs": 1}
    target_query = urllib.parse.urlencode(query)
    fallback_dirfs_path = f"local?{target_query}"
    target = Target.open(fallback_dirfs_path)

    assert "fallback-to-directory-fs" in target.path_query

    # no volumes should be attached, since no disks were attached
    # see comment in LocalLoader.map() for more info
    assert len(target.volumes) == 0

    assert len(target.fs.mounts) == 1
    assert target.fs.mounts.get("/")
    assert isinstance(target.fs.mounts["/"], DirectoryFilesystem)


def test_target__generic_name_no_path_no_os() -> None:
    target_bare = Target()

    assert target_bare._generic_name == "Unknown"


def test_target__generic_name_no_path_with_os() -> None:
    target_bare = Target()
    target_bare.os = "test"

    assert target_bare._generic_name == "Unknown-test"


def test_target__generic_name_with_path(target_bare: Target) -> None:
    assert target_bare._generic_name == target_bare.path.name


def test_target_name_no_hostname() -> None:
    test_name = "test-target"

    with patch("dissect.target.target.Target._generic_name", PropertyMock(return_value=test_name)):
        target_bare = Target()
        target_bare.hostname = None

        assert target_bare.name == test_name


def test_target_name_hostname_raises() -> None:
    test_name = "test-target"

    with (
        patch("dissect.target.target.Target._generic_name", PropertyMock(return_value=test_name)),
        patch("dissect.target.target.Target.hostname", PropertyMock(side_effect=Exception("ERROR")), create=True),
    ):
        target_bare = Target()

        assert target_bare.name == test_name


def test_target_name_set() -> None:
    test_name = "test-target"
    target_bare = Target()
    target_bare._name = test_name

    assert target_bare.name == test_name


def test_target_name_target_applied() -> None:
    test_name = "test-target"

    with patch("dissect.target.target.Target._generic_name", PropertyMock(return_value=test_name)):
        target_bare = Target()
        target_bare.hostname = None
        target_bare._applied = True

        assert target_bare.name == test_name
        assert target_bare._name == test_name


def test_target_set_event_callback() -> None:
    mock_callback1 = MagicMock()
    mock_callback2 = MagicMock()

    class MockTarget(Target):
        event_callbacks: ClassVar = defaultdict(set)

    MockTarget.set_event_callback(event_type=None, event_callback=mock_callback1)
    MockTarget.set_event_callback(event_type=None, event_callback=mock_callback2)
    MockTarget.set_event_callback(event_type=Event.FUNC_EXEC, event_callback=mock_callback1)
    MockTarget.set_event_callback(event_type=Event.FUNC_EXEC, event_callback=mock_callback2)

    assert hasattr(MockTarget, "event_callbacks")

    event_callbacks = MockTarget.event_callbacks

    assert None in event_callbacks
    assert Event.FUNC_EXEC in event_callbacks

    assert len(event_callbacks[None]) == 2
    assert len(event_callbacks[Event.FUNC_EXEC]) == 2

    assert mock_callback1 in event_callbacks[None]
    assert mock_callback2 in event_callbacks[None]
    assert mock_callback1 in event_callbacks[Event.FUNC_EXEC]
    assert mock_callback2 in event_callbacks[Event.FUNC_EXEC]


def test_target_send_event() -> None:
    mock_callback1 = MagicMock()
    mock_callback2 = MagicMock()

    class MockTarget(Target):
        event_callbacks: ClassVar = defaultdict(set)

    MockTarget.set_event_callback(event_type=None, event_callback=mock_callback1)
    MockTarget.set_event_callback(event_type=Event.FUNC_EXEC, event_callback=mock_callback2)

    target_bare = MockTarget()
    target_bare.send_event(None, test="None")
    target_bare.send_event(Event.FUNC_EXEC, test="FUNC_EXEC")

    calls = [
        call(target_bare, None, test="None"),
        call(target_bare, Event.FUNC_EXEC, test="FUNC_EXEC"),
    ]
    mock_callback1.assert_has_calls(calls)
    mock_callback2.assert_called_once_with(target_bare, Event.FUNC_EXEC, test="FUNC_EXEC")


def test_empty_vs(target_bare: Target) -> None:
    mock_disk = MagicMock()
    mock_disk.vs = None

    mock_volume_system = MagicMock()
    mock_volume_system.volumes = []

    target_bare.disks.add(mock_disk)

    with patch("dissect.target.volume.open") as volume_open:
        volume_open.return_value = mock_volume_system

        target_bare.disks.apply()

        assert len(target_bare.volumes) == 1
        assert target_bare.volumes[0].disk is mock_disk
        assert target_bare.volumes[0].offset == 0


def test_nested_vs(target_bare: Target) -> None:
    mock_base_volume = MagicMock()
    mock_base_volume.offset = 0
    mock_base_volume.vs = None
    mock_base_volume.fs = None
    target_bare.volumes.add(mock_base_volume)

    mock_volume = MagicMock()
    mock_volume.offset = 0
    mock_volume.vs.__type__ = "disk"
    mock_volume.fs = None

    mock_volume_system = MagicMock()
    mock_volume_system.volumes = [mock_volume]

    with patch("dissect.target.volume.open") as volume_open, patch("dissect.target.filesystem.open") as filesystem_open:
        volume_open.return_value = mock_volume_system

        target_bare.volumes.apply()
        filesystem_open.assert_has_calls([call(mock_base_volume), call(mock_volume)])
        assert len(target_bare.volumes) == 2
        assert len(target_bare.filesystems) == 2


def test_vs_offset_0(target_bare: Target) -> None:
    mock_disk = MagicMock()
    mock_disk.vs = None
    target_bare.disks.add(mock_disk)

    mock_volume = MagicMock()
    mock_volume.offset = 0
    mock_volume.vs.__type__ = "disk"
    mock_volume.fs = None

    mock_volume_system = MagicMock()
    mock_volume_system.volumes = [mock_volume]

    with patch("dissect.target.volume.open") as volume_open, patch("dissect.target.filesystem.open") as filesystem_open:
        volume_open.return_value = mock_volume_system

        target_bare.disks.apply()
        volume_open.assert_called_once_with(mock_disk)
        assert len(target_bare.volumes) == 1

        target_bare.volumes.apply()
        # volume.open must still only be called once
        volume_open.assert_called_once_with(mock_disk)
        filesystem_open.assert_called_once_with(mock_volume)
        assert len(target_bare.volumes) == 1
        assert len(target_bare.filesystems) == 1


def test_empty_volume_log(target_bare: Target, caplog: pytest.LogCaptureFixture) -> None:
    """Test whether we correctly log an undetectable filesystem warning."""
    mock_volume = MagicMock()
    mock_volume.fs = None
    mock_volume.read = Mock(side_effect=lambda size: b"\x00" * size)

    target_bare.volumes.add(mock_volume)

    with caplog.at_level(logging.DEBUG, target_bare.log.name):
        target_bare.volumes.apply()

    assert "Skipping empty volume" in caplog.text
    assert "Can't identify filesystem" not in caplog.text

    # Now test that we do log a warning if the volume is not empty
    mock_volume.read = Mock(side_effect=lambda size: b"\x01" * size)

    caplog.clear()
    with caplog.at_level(logging.DEBUG, target_bare.log.name):
        target_bare.volumes.apply()

    assert "Skipping empty volume" not in caplog.text
    assert "Can't identify filesystem" in caplog.text


@pytest.mark.parametrize("nr_of_fs", [1, 2])
def test_fs_mount_others(target_unix: Target, nr_of_fs: int) -> None:
    for _ in range(nr_of_fs):
        target_unix.filesystems.add(Mock())

    target_unix._mount_others()

    for idx in range(nr_of_fs):
        assert f"/$fs$/fs{idx}" in target_unix.fs.mounts
        assert target_unix.fs.path(f"$fs$/fs{idx}").exists()

    assert not target_unix.fs.path(f"$fs$/fs{nr_of_fs}").exists()


@pytest.mark.parametrize("nr_of_fs", [1, 2])
def test_fs_mount_already_there(target_unix: Target, nr_of_fs: int) -> None:
    for idx in range(nr_of_fs):
        target_unix.filesystems.add(Mock())
        target_unix._mount_others()

        assert f"/$fs$/fs{idx}" in target_unix.fs.mounts
        assert target_unix.fs.path(f"$fs$/fs{idx}").exists()


def test_children_on_invalid_target(caplog: pytest.LogCaptureFixture, tmp_path: Path) -> None:
    """Test that we don't attempt to load child targets on an invalid target."""
    p = tmp_path.joinpath("empty-dir")
    p.mkdir()

    mock_loader = Mock()
    mock_loader.find_all.return_value = [None]
    mock_loader.return_value = None

    with patch.object(loader, "find_loader", return_value=mock_loader):
        try:
            list(Target.open_all([p], include_children=True))
        except Exception:
            pass

    assert "Failed to load child target from None" not in caplog.text


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_open_uri(opener: Callable[[str | Path], Target]) -> None:
    """Test that we can open a URI with a custom scheme."""

    class MockLoader(loader.Loader):
        def __init__(self, path: Path, parsed_path: urllib.parse.ParseResult | None = None):
            super().__init__(path, parsed_path=parsed_path, resolve=False)

        def map(self, target: Target) -> None:
            target.filesystems.add(VirtualFilesystem())

    with patch.dict(loader.LOADERS_BY_SCHEME, {"mock": MockLoader}):
        path = "mock://user:password@example.com:1337/path/to/resource?query=1&other=2"
        target = opener(path)

        assert target.path == Path("example.com:1337/path/to/resource")
        assert target.path_query == {"query": "1", "other": "2"}
        assert isinstance(target._loader, MockLoader)
        assert target._loader.parsed_path
        assert target._loader.parsed_path.scheme == "mock"
        assert target._loader.parsed_path.netloc == "user:password@example.com:1337"
        assert target._loader.parsed_path.path == "/path/to/resource"
        assert target._loader.parsed_path.query == "query=1&other=2"


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
@pytest.mark.skipif(platform.system() == "Windows", reason="Unix-specific test")
def test_open_uri_path(opener: Callable[[str | Path], Target], tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    (tmp_path / "mock:").mkdir()
    (tmp_path / "mock:" / "file").write_bytes(b"I'm a hard drive!")

    monkeypatch.chdir(tmp_path)

    # If we pass in a Path object, we should not parse it as a URI
    target = opener(Path("mock:/file"))
    assert target.path == Path("mock:/file")
    assert isinstance(target._loader, RawLoader)
    assert target._loader.path == Path("mock:/file")
    assert target._loader.parsed_path is None

    # If we pass it in as a string and we have no loader for it, it should still be treated as a local path
    target = opener("mock:/file")
    assert target.path == Path("mock:/file")
    assert isinstance(target._loader, RawLoader)
    assert target._loader.path == Path("mock:/file")
    assert target._loader.parsed_path is None


@pytest.mark.parametrize(
    ("path", "expected"),
    [
        pytest.param("", Path(), id="str-empty"),
        # Empty path with query, could be the filename
        pytest.param("?query=1", Path("?query=1"), id="str-empty-query"),
        pytest.param("who-am-i?", Path("who-am-i?"), id="str-question-mark"),
        # Normal relative file paths
        pytest.param("Test", Path("Test"), id="str-relative"),
        pytest.param("/Test", Path("/Test"), id="str-absolute"),
        pytest.param("./Test", Path("./Test"), id="str-explicit-relative"),
        # Windows drive letter path
        pytest.param("C:\\Test", Path("C:\\Test"), id="str-drive-letter"),
        # The "local" path is a special case because we use it for the local loader, but it should behave the same
        pytest.param("local", Path("local"), id="str-local"),
        # local path with query, still should be treated the same for consistency
        pytest.param("local?query=1&other=2", Path("local?query=1&other=2"), id="str-local-query"),
        # Normal absolute paths
        pytest.param("/path/to/Test", Path("/path/to/Test"), id="str-absolute-directory"),
        pytest.param(
            "/path/to/Test?query=1&other=2", Path("/path/to/Test?query=1&other=2"), id="str-absolute-directory-query"
        ),
        pytest.param("C:\\path\\to\\Test", Path("C:\\path\\to\\Test"), id="str-drive-letter-directory"),
        pytest.param("\\\\UNC\\share\\path\\to\\Test", Path("\\\\UNC\\share\\path\\to\\Test"), id="str-unc"),
        # URI paths is where we start to see differences
        pytest.param("uri://Example.com/path/to/Test", Path("Example.com/path/to/Test"), id="str-uri"),
        pytest.param(
            "uri://wing@SafeComputer.com:1337/path/to/Test",
            Path("SafeComputer.com:1337/path/to/Test"),
            id="str-uri-user",
        ),
        pytest.param(
            "uri://user:password@Example.com:6969/path/to/Test",
            Path("Example.com:6969/path/to/Test"),
            id="str-uri-user-password",
        ),
        pytest.param(
            "uri://Example.com/path/to/Test?query=1&other=2", Path("Example.com/path/to/Test"), id="str-uri-query"
        ),
        # Path objects should behave the same as strings, except that the type should be preserved
        pytest.param(Path(), Path(), id="path-empty"),
        pytest.param(Path("?query=1"), Path("?query=1"), id="path-empty-query"),
        pytest.param(Path("who-am-i?"), Path("who-am-i?"), id="path-question-mark"),
        pytest.param(Path("Test"), Path("Test"), id="path-relative"),
        pytest.param(Path("/Test"), Path("/Test"), id="path-absolute"),
        pytest.param(Path("./Test"), Path("./Test"), id="path-explicit-relative"),
        pytest.param(Path("C:\\Test"), Path("C:\\Test"), id="path-drive-letter"),
        pytest.param(Path("local"), Path("local"), id="path-local"),
        pytest.param(Path("local?query=1&other=2"), Path("local?query=1&other=2"), id="path-local-query"),
        pytest.param(Path("\\\\UNC\\share\\path\\to\\Test"), Path("\\\\UNC\\share\\path\\to\\Test"), id="path-unc"),
        pytest.param(TargetPath(VirtualFilesystem(), ""), TargetPath(VirtualFilesystem(), ""), id="path-custom-empty"),
        pytest.param(
            TargetPath(VirtualFilesystem(), "Custom"),
            TargetPath(VirtualFilesystem(), "Custom"),
            id="path-custom-simple",
        ),
        pytest.param(
            TargetPath(VirtualFilesystem(), "/sub/Custom"),
            TargetPath(VirtualFilesystem(), "/sub/Custom"),
            id="path-custom-directory",
        ),
    ],
)
def test_expected_path(path: str | Path, expected: Path) -> None:
    """Test that ``target.path`` has the expected value for a given path."""
    target = Target(path)

    assert target.path == expected
    if isinstance(path, Path):
        assert type(target.path) is type(expected)



from __future__ import annotations

import hashlib
from typing import TYPE_CHECKING

from Crypto.Hash import MD4

from dissect.target.helpers import keychain
from dissect.target.plugins.os.windows.credential.credhist import (
    CredHistFile,
    CredHistPlugin,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_credhist() -> None:
    """The provided CREDHIST file has the following password history: ``user -> password -> password3``.
    The current password of the user is ``password4``.
    """
    with absolute_path("_data/plugins/os/windows/credhist/CREDHIST").open("rb") as fh:
        ch = CredHistFile(fh)

    assert len(ch.entries) == 3

    for entry in ch.entries:
        assert ch.entries[0].version == 1
        assert entry.user_sid.upper() == "S-1-5-21-1342509979-482553916-3960431919-1000"

    ch.decrypt(password_hash=sha1("password4"))

    assert str(ch.entries[0].guid) == "99ec7176-d16c-41bd-9c94-d3a4c5b94232"
    assert ch.entries[0].sha1 == sha1("user")
    assert ch.entries[0].nt == md4("user")

    assert str(ch.entries[1].guid) == "120a3a30-309c-4fda-bfb8-06f44ea93cb2"
    assert ch.entries[1].sha1 == sha1("password")
    assert ch.entries[1].nt == md4("password")

    assert str(ch.entries[2].guid) == "5657891f-28dd-4f69-baba-95e44bcd178a"
    assert ch.entries[2].sha1 == sha1("password3")
    assert ch.entries[2].nt == md4("password3")


def test_credhist_partial(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    """Test if we can get a partially decrypted CREDHIST chain if we know an intermediate password.

    The latest entry is encrypted with 'password4' but we provide 'password3'. The plugin
    should decrypt every entry except the latest entry.
    """
    fs_win.map_file(
        "Users/John/AppData/Roaming/Microsoft/Protect/CREDHIST",
        absolute_path("_data/plugins/os/windows/credhist/CREDHIST"),
    )
    target_win_users.add_plugin(CredHistPlugin)

    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value="password3",
        identifier=None,
        provider="user",
    )

    results = list(target_win_users.credhist())
    assert len(results) == 3
    assert [result.nt for result in results] == [md4("user").hex(), md4("password").hex(), None]


def md4(plaintext: str) -> str:
    return MD4.new(plaintext.encode("utf-16-le")).digest()


def sha1(plaintext: str) -> str:
    return hashlib.sha1(plaintext.encode("utf-16-le")).digest()



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.archive import vma

from dissect.target.containers.raw import RawContainer
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class VmaLoader(Loader):
    """Load Proxmox Virtual Machine Archive (VMA) files.

    References:
        - https://pve.proxmox.com/wiki/VMA
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        self.vma = vma.VMA(path.open("rb"))

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".vma"

    def map(self, target: Target) -> None:
        for device in self.vma.devices():
            target.disks.add(RawContainer(device.open()))



from __future__ import annotations

import base64
import hmac
import json
import logging
from base64 import b64decode
from hashlib import pbkdf2_hmac, sha1
from itertools import chain
from typing import TYPE_CHECKING

from dissect.sql import sqlite3
from dissect.sql.exceptions import Error as SQLError
from dissect.util.ts import from_unix_ms, from_unix_us

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.apps.browser.browser import (
    GENERIC_COOKIE_FIELDS,
    GENERIC_DOWNLOAD_RECORD_FIELDS,
    GENERIC_EXTENSION_RECORD_FIELDS,
    GENERIC_HISTORY_RECORD_FIELDS,
    GENERIC_PASSWORD_RECORD_FIELDS,
    BrowserPlugin,
    try_idna,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.sql.sqlite3 import SQLite3

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.plugins.general.users import UserRecord
    from dissect.target.target import Target

try:
    from asn1crypto import algos, core

    HAS_ASN1 = True

except ImportError:
    HAS_ASN1 = False


try:
    from Crypto.Cipher import AES, DES3
    from Crypto.Util.Padding import unpad

    HAS_CRYPTO = True

except ImportError:
    HAS_CRYPTO = False

FIREFOX_EXTENSION_RECORD_FIELDS = [
    ("uri", "source_uri"),
    ("string[]", "optional_permissions"),
]

log = logging.getLogger(__name__)


class FirefoxPlugin(BrowserPlugin):
    """Firefox browser plugin."""

    __namespace__ = "firefox"

    USER_DIRS = (
        # Windows
        "AppData/Roaming/Mozilla/Firefox/Profiles",
        "AppData/local/Mozilla/Firefox/Profiles",
        # Linux
        ".mozilla/firefox",
        "snap/firefox/common/.mozilla/firefox",
        ".var/app/org.mozilla.firefox/.mozilla/firefox",
        # macOS
        "Library/Application Support/Firefox",
    )

    SYSTEM_DIRS = ("/data/data/org.mozilla.vrbrowser/files/mozilla",)

    BrowserHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/firefox/history", GENERIC_HISTORY_RECORD_FIELDS
    )

    BrowserCookieRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/firefox/cookie", GENERIC_COOKIE_FIELDS
    )

    BrowserDownloadRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/firefox/download", GENERIC_DOWNLOAD_RECORD_FIELDS
    )

    BrowserExtensionRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/firefox/extension",
        GENERIC_EXTENSION_RECORD_FIELDS + FIREFOX_EXTENSION_RECORD_FIELDS,
    )

    BrowserPasswordRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/firefox/password", GENERIC_PASSWORD_RECORD_FIELDS
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.dirs: list[tuple[UserRecord, TargetPath]] = []

        for user_details in self.target.user_details.all_with_home():
            for directory in self.USER_DIRS:
                cur_dir = user_details.home_path.joinpath(directory)
                if not cur_dir.exists():
                    continue
                self.dirs.append((user_details.user, cur_dir))

        for directory in self.SYSTEM_DIRS:
            if (cur_dir := target.fs.path(directory)).exists():
                self.dirs.append((None, cur_dir))

    def check_compatible(self) -> None:
        if not len(self.dirs):
            raise UnsupportedPluginError("No Firefox directories found")

    def _iter_profiles(self) -> Iterator[tuple[UserRecord, TargetPath, TargetPath]]:
        """Yield user directories."""
        for user, cur_dir in self.dirs:
            for profile_dir in cur_dir.iterdir():
                if not profile_dir.is_dir():
                    continue
                yield user, cur_dir, profile_dir

    def _iter_db(self, filename: str) -> Iterator[tuple[UserRecord, SQLite3]]:
        """Yield opened history database files of all users.

        Args:
            filename: The filename of the database.

        Yields:
            Opened SQLite3 databases.
        """
        iter_system = ((None, system_dir, None) for user, system_dir in self.dirs if user is None)

        for user, cur_dir, profile_dir in chain(iter_system, self._iter_profiles()):
            if user is None and profile_dir is None:
                db_file = cur_dir.parent.joinpath(filename)
                # On some Android variants, some files may exist in the base directory (places.sqlite) but others
                # in a nested profile directory (cookies.sqlite)
                # /data/data/org.mozilla.vrbrowser/files/places.sqlite
                # /data/data/org.mozilla.vrbrowser/files/mozilla/xxxxxx.default/cookies.sqlite
                if not db_file.exists():
                    continue
            else:
                db_file = profile_dir.joinpath(filename)

            try:
                yield user, db_file, sqlite3.SQLite3(db_file.open())
            except FileNotFoundError:
                self.target.log.info("Could not find %s file: %s", filename, db_file)
            except SQLError as e:
                self.target.log.warning("Could not open %s file: %s", filename, db_file)
                self.target.log.debug("", exc_info=e)

    @export(record=BrowserHistoryRecord)
    def history(self) -> Iterator[BrowserHistoryRecord]:
        """Return browser history records from Firefox.

        Yields BrowserHistoryRecord with the following fields:

        .. code-block:: text

            ts (datetime): Visit timestamp.
            browser (string): The browser from which the records are generated from.
            id (string): Record ID.
            url (uri): History URL.
            title (string): Page title.
            description (string): Page description.
            host (string): Hostname.
            visit_type (varint): Visit type.
            visit_count (varint): Amount of visits.
            hidden (string): Hidden value.
            typed (boolean): Typed value.
            session (varint): Session value.
            from_visit (varint): Record ID of the "from" visit.
            from_url (uri): URL of the "from" visit.
            source: (path): The source file of the history record.
        """
        from_timestamp = from_unix_ms if self.target.os == OperatingSystem.ANDROID else from_unix_us

        for user, db_file, db in self._iter_db("places.sqlite"):
            try:
                places = {row.id: row for row in db.table("moz_places").rows()}
                visits = {}

                for row in db.table("moz_historyvisits").rows():
                    visits[row.id] = row
                    place = places[row.place_id]

                    if row.from_visit and row.from_visit in visits:
                        from_visit = visits[row.from_visit]
                        from_place = places[from_visit.place_id]
                    else:
                        from_visit, from_place = None, None

                    yield self.BrowserHistoryRecord(
                        ts=from_timestamp(row.visit_date),
                        browser="firefox",
                        id=row.id,
                        url=try_idna(place.url),
                        title=place.title,
                        description=place.description,
                        host="".join(list(reversed(try_idna(place.rev_host).decode()))).lstrip(".")
                        if place.rev_host
                        else None,
                        visit_type=row.visit_type,
                        visit_count=place.visit_count,
                        hidden=place.hidden,
                        typed=place.typed,
                        session=row.session,
                        from_visit=row.from_visit or None,
                        from_url=try_idna(from_place.url) if from_place else None,
                        source=db_file,
                        _target=self.target,
                        _user=user,
                    )
            except SQLError as e:  # noqa: PERF203
                self.target.log.warning("Error processing history file: %s", db_file)
                self.target.log.debug("", exc_info=e)

    @export(record=BrowserCookieRecord)
    def cookies(self) -> Iterator[BrowserCookieRecord]:
        """Return browser cookie records from Firefox.

        Args:
            browser_name: The name of the browser as a string.

        Yields:

        .. code-block:: text

            Records with the following fields:
                ts_created (datetime): Cookie created timestamp.
                ts_last_accessed (datetime): Cookie last accessed timestamp.
                browser (string): The browser from which the records are generated from.
                name (string): The cookie name.
                value (string): The cookie value.
                host (string): Cookie host key.
                path (string): Cookie path.
                expiry (varint): Cookie expiry.
                is_secure (bool): Cookie secury flag.
                is_http_only (bool): Cookie http only flag.
                same_site (bool): Cookie same site flag.
        """
        for user, db_file, db in self._iter_db("cookies.sqlite"):
            try:
                for cookie in db.table("moz_cookies").rows():
                    yield self.BrowserCookieRecord(
                        ts_created=from_unix_us(cookie.creationTime),
                        ts_last_accessed=from_unix_us(cookie.lastAccessed),
                        browser="firefox",
                        name=cookie.name,
                        value=cookie.value,
                        host=cookie.host,
                        path=cookie.path,
                        expiry=cookie.expiry,
                        is_secure=bool(cookie.isSecure),
                        is_http_only=bool(cookie.isHttpOnly),
                        same_site=bool(cookie.sameSite),
                        source=db_file,
                        _target=self.target,
                        _user=user,
                    )
            except SQLError as e:  # noqa: PERF203
                self.target.log.warning("Error processing cookie file: %s", db_file)
                self.target.log.debug("", exc_info=e)

    @export(record=BrowserDownloadRecord)
    def downloads(self) -> Iterator[BrowserDownloadRecord]:
        """Return browser download records from Firefox.

        Yields BrowserDownloadRecord with the following fields:

        .. code-block:: text

            ts_start (datetime): Download start timestamp.
            ts_end (datetime): Download end timestamp.
            browser (string): The browser from which the records are generated from.
            id (string): Record ID.
            path (string): Download path.
            url (uri): Download URL.
            size (varint): Download file size.
            state (varint): Download state number.
            source: (path): The source file of the download record.
        """
        for user, db_file, db in self._iter_db("places.sqlite"):
            try:
                places = {row.id: row for row in db.table("moz_places").rows()}
                if not (moz_anno_attributes := db.table("moz_anno_attributes")):
                    continue

                attributes = {row.id: row.name for row in moz_anno_attributes.rows()}
                annotations = {}

                for row in db.table("moz_annos"):
                    attribute_name = attributes.get(row.anno_attribute_id, row.anno_attribute_id)

                    content = json.loads(row.content) if attribute_name == "downloads/metaData" else row.content

                    if row.place_id not in annotations:
                        annotations[row.place_id] = {"id": row.id}

                    annotations[row.place_id][attribute_name] = {
                        "content": content,
                        "flags": row.flags,
                        "expiration": row.expiration,
                        "type": row.type,
                        "date_added": from_unix_us(row.dateAdded),
                        "last_modified": from_unix_us(row.lastModified),
                    }

                for place_id, annotation in annotations.items():
                    if "downloads/metaData" not in annotation:
                        continue

                    metadata = annotation.get("downloads/metaData", {})

                    ts_end = None
                    size = None
                    state = None

                    content = metadata.get("content")
                    if content:
                        ts_end = metadata.get("content").get("endTime")
                        ts_end = from_unix_ms(ts_end) if ts_end else None

                        size = content.get("fileSize")
                        state = content.get("state")

                    dest_file_info = annotation.get("downloads/destinationFileURI", {})

                    if download_path := dest_file_info.get("content"):
                        download_path = self.target.fs.path(download_path)

                    place = places.get(place_id)
                    url = place.get("url")
                    url = try_idna(url) if url else None

                    yield self.BrowserDownloadRecord(
                        ts_start=dest_file_info.get("date_added"),
                        ts_end=ts_end,
                        browser="firefox",
                        id=annotation.get("id"),
                        path=download_path,
                        url=url,
                        size=size,
                        state=state,
                        source=db_file,
                        _target=self.target,
                        _user=user,
                    )
            except SQLError as e:
                self.target.log.warning("Error processing history file: %s", db_file)
                self.target.log.debug("", exc_info=e)

    @export(record=BrowserExtensionRecord)
    def extensions(self) -> Iterator[BrowserExtensionRecord]:
        """Return browser extension records for Firefox.

        Yields BrowserExtensionRecord with the following fields:

        .. code-block:: text

            ts_install (datetime): Extension install timestamp.
            ts_update (datetime): Extension update timestamp.
            browser (string): The browser from which the records are generated.
            id (string): Extension unique identifier.
            name (string): Name of the extension.
            short_name (string): Short name of the extension.
            default_title (string): Default title of the extension.
            description (string): Description of the extension.
            version (string): Version of the extension.
            ext_path (path): Relative path of the extension.
            from_webstore (boolean): Extension from webstore.
            permissions (string[]): Permissions of the extension.
            manifest (varint): Version of the extensions' manifest.
            optional_permissions (string[]): Optional permissions of the extension.
            source_uri (path): Source path from which the extension was downloaded.
            source (path): The source file of the download record.
        """
        for user, _, profile_dir in self._iter_profiles():
            extension_file = profile_dir.joinpath("extensions.json")

            if not extension_file.exists():
                self.target.log.warning(
                    "No 'extensions.json' addon file found for user %s in directory %s",
                    user.name,
                    profile_dir,
                )
                continue

            try:
                extensions = json.load(extension_file.open())

                for extension in extensions.get("addons", []):
                    yield self.BrowserExtensionRecord(
                        ts_install=from_unix_ms(extension.get("installDate", 0)),
                        ts_update=from_unix_ms(extension.get("updateDate", 0)),
                        browser="firefox",
                        extension_id=extension.get("id"),
                        name=(extension.get("defaultLocale", {}) or {}).get("name"),
                        short_name=None,
                        default_title=None,
                        description=(extension.get("defaultLocale", {}) or {}).get("description"),
                        version=extension.get("version"),
                        ext_path=extension.get("path"),
                        from_webstore=None,
                        permissions=(extension.get("userPermissions", {}) or {}).get("permissions"),
                        manifest_version=extension.get("manifestVersion"),
                        source_uri=extension.get("sourceURI"),
                        optional_permissions=(extension.get("optionalPermissions", {}) or {}).get("permissions"),
                        source=extension_file,
                        _target=self.target,
                        _user=user,
                    )

            except FileNotFoundError:
                self.target.log.info(
                    "No 'extensions.json' addon file found for user %s in directory %s",
                    user.name,
                    profile_dir,
                )
            except json.JSONDecodeError:
                self.target.log.warning(
                    "extensions.json file in directory %s is malformed, consider inspecting the file manually",
                    profile_dir,
                )

    @export(record=BrowserPasswordRecord)
    def passwords(self) -> Iterator[BrowserPasswordRecord]:
        """Return Firefox browser password records.

        Automatically decrypts passwords from Firefox 58 onwards (2018) if no primary password is set.
        Alternatively, you can supply a primary password through the keychain to access the Firefox password store.

        ``PASSPHRASE`` passwords in the keychain with providers ``browser``, ``firefox``, ``user`` and no provider
        can be used to decrypt secrets for this plugin.

        References:
            - https://github.com/lclevy/firepwd
        """
        for user, _, profile_dir in self._iter_profiles():
            login_file = profile_dir.joinpath("logins.json")
            key3_file = profile_dir.joinpath("key3.db")
            key4_file = profile_dir.joinpath("key4.db")

            if not login_file.exists():
                self.target.log.warning(
                    "No 'logins.json' password file found for user %s in directory %s",
                    user.name,
                    profile_dir,
                )
                continue

            if key3_file.exists() and not key4_file.exists():
                self.target.log.warning("Unsupported file 'key3.db' found in %s", profile_dir)
                continue

            if not key4_file.exists():
                self.target.log.warning("No 'key4.db' found in %s", profile_dir)
                continue

            try:
                logins = json.load(login_file.open())

                for login in logins.get("logins", []):
                    decrypted_username = None
                    decrypted_password = None

                    for password in self.keychain():
                        try:
                            decrypted_username, decrypted_password = decrypt(
                                login.get("encryptedUsername"),
                                login.get("encryptedPassword"),
                                key4_file,
                                password,
                            )
                        except ValueError as e:
                            self.target.log.warning("Exception while trying to decrypt")
                            self.target.log.debug("", exc_info=e)

                        if decrypted_password and decrypted_username:
                            break

                    yield self.BrowserPasswordRecord(
                        ts_created=from_unix_ms(login.get("timeCreated", 0)),
                        ts_last_used=from_unix_ms(login.get("timeLastUsed", 0)),
                        ts_last_changed=from_unix_ms(login.get("timePasswordChanged", 0)),
                        browser="firefox",
                        id=login.get("id"),
                        url=login.get("hostname"),
                        encrypted_username=base64.b64decode(login.get("encryptedUsername", "")),
                        encrypted_password=base64.b64decode(login.get("encryptedPassword", "")),
                        decrypted_username=decrypted_username,
                        decrypted_password=decrypted_password,
                        source=login_file,
                        _target=self.target,
                        _user=user,
                    )

            except FileNotFoundError:
                self.target.log.info(
                    "No password file found for user %s in directory %s",
                    user.name,
                    profile_dir,
                )
            except json.JSONDecodeError:
                self.target.log.warning(
                    "logins.json file in directory %s is malformed, consider inspecting the file manually",
                    profile_dir,
                )


# Define separately because it is not defined in asn1crypto
pbeWithSha1AndTripleDES_CBC = "1.2.840.113549.1.12.5.1.3"
CKA_ID = b"\xf8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"


def decrypt_moz_3des(global_salt: bytes, primary_password: bytes, entry_salt: str, encrypted: bytes) -> bytes:
    if not HAS_CRYPTO:
        raise ValueError("Missing pycryptodome dependency")

    hp = sha1(global_salt + primary_password).digest()
    pes = entry_salt + b"\x00" * (20 - len(entry_salt))
    chp = sha1(hp + entry_salt).digest()
    k1 = hmac.new(chp, pes + entry_salt, sha1).digest()
    tk = hmac.new(chp, pes, sha1).digest()
    k2 = hmac.new(chp, tk + entry_salt, sha1).digest()
    k = k1 + k2
    iv = k[-8:]
    key = k[:24]
    return DES3.new(key, DES3.MODE_CBC, iv).decrypt(encrypted)


def decode_login_data(data: str) -> tuple[bytes, bytes, bytes]:
    """Decode Firefox login data.

    Args:
        data: Base64 encoded data in string format.

    Raises:
        ValueError: When missing ``pycryptodome`` or ``asn1crypto`` dependencies.

    Returns:
        Tuple of bytes with ``key_id``, ``iv`` and ``ciphertext``
    """

    # SEQUENCE {
    #     KEY_ID
    #     SEQUENCE {
    #         OBJECT_IDENTIFIER
    #         IV
    #     }
    #     CIPHERTEXT
    # }

    if not HAS_CRYPTO:
        raise ValueError("Missing pycryptodome dependency")

    if not HAS_ASN1:
        raise ValueError("Missing asn1crypto dependency")

    decoded = core.load(b64decode(data))
    key_id = decoded[0].native
    iv = decoded[1][1].native
    ciphertext = decoded[2].native
    return key_id, iv, ciphertext


def decrypt_pbes2(decoded_item: core.Sequence, primary_password: bytes, global_salt: bytes) -> bytes:
    """Decrypt an item with the given primary password and salt.

    Args:
        decoded_item: ``core.Sequence`` is a ``list`` representation of ``SEQUENCE`` as described below.
        primary_password: ``bytes`` of Firefox primary password to decrypt ciphertext with.
        global_salt: ``bytes`` of salt to prepend to primary password when calculating AES key.

    Raises:
        ValueError: When missing ``pycryptodome`` or ``asn1crypto`` dependencies.

    Returns:
        Bytes of decrypted AES ciphertext.
    """

    # SEQUENCE {
    #   SEQUENCE {
    #     OBJECTIDENTIFIER 1.2.840.113549.1.5.13 => pkcs5 pbes2
    #     SEQUENCE {
    #       SEQUENCE {
    #         OBJECTIDENTIFIER 1.2.840.113549.1.5.12 => pbkdf2
    #         SEQUENCE {
    #           OCTETSTRING 32 bytes, entrySalt
    #           INTEGER 01
    #           INTEGER 20
    #           SEQUENCE {
    #             OBJECTIDENTIFIER 1.2.840.113549.2.9 => hmacWithSHA256
    #           }
    #         }
    #       }
    #       SEQUENCE {
    #         OBJECTIDENTIFIER 2.16.840.1.101.3.4.1.42 => aes256-CBC
    #         OCTETSTRING 14 bytes, iv
    #       }
    #     }
    #   }
    #   OCTETSTRING encrypted
    # }

    if not HAS_CRYPTO:
        raise ValueError("Missing pycryptodome dependency")

    if not HAS_ASN1:
        raise ValueError("Missing asn1crypto dependency")

    pkcs5_oid = decoded_item[0][1][0][0].dotted
    if algos.KdfAlgorithmId.map(pkcs5_oid) != "pbkdf2":
        raise ValueError(f"Expected pbkdf2 object identifier, got: {pkcs5_oid}")

    sha256_oid = decoded_item[0][1][0][1][3][0].dotted
    if algos.HmacAlgorithmId.map(sha256_oid) != "sha256":
        raise ValueError(f"Expected SHA256 object identifier, got: {pkcs5_oid}")

    aes256_cbc_oid = decoded_item[0][1][1][0].dotted
    if algos.EncryptionAlgorithmId.map(aes256_cbc_oid) != "aes256_cbc":
        raise ValueError(f"Expected AES256-CBC object identifier, got: {pkcs5_oid}")

    entry_salt = decoded_item[0][1][0][1][0].native
    iteration_count = decoded_item[0][1][0][1][1].native
    key_length = decoded_item[0][1][0][1][2].native

    if key_length != 32:
        raise ValueError(f"Expected key_length to be 32, got: {key_length}")

    k = sha1(global_salt + primary_password).digest()
    key = pbkdf2_hmac("sha256", k, entry_salt, iteration_count, dklen=key_length)

    iv = b"\x04\x0e" + decoded_item[0][1][1][1].native
    cipher_text = decoded_item[1].native
    return AES.new(key, AES.MODE_CBC, iv).decrypt(cipher_text)


def decrypt_sha1_triple_des_cbc(decoded_item: core.Sequence, primary_password: bytes, global_salt: bytes) -> bytes:
    """Decrypt an item with the given Firefox primary password and salt.

    Args:
        decoded_item: ``core.Sequence`` is a ``list`` representation of ``SEQUENCE`` as described below.
        primary_password: ``bytes`` of Firefox primary password to decrypt ciphertext with.
        global_salt: ``bytes`` of salt to prepend to primary password when calculating AES key.

    Raises:
        ValueError: When missing ``pycryptodome`` or ``asn1crypto`` dependencies.

    Returns:
        Bytes of decrypted 3DES ciphertext.
    """

    # SEQUENCE {
    #     SEQUENCE {
    #         OBJECTIDENTIFIER 1.2.840.113549.1.12.5.1.3
    #         SEQUENCE {
    #             OCTETSTRING entry_salt
    #             INTEGER 01
    #         }
    #     }
    #     OCTETSTRING encrypted
    # }

    entry_salt = decoded_item[0][1][0].native
    cipher_text = decoded_item[1].native
    key = decrypt_moz_3des(global_salt, primary_password, entry_salt, cipher_text)
    return key[:24]


def decrypt_master_key(decoded_item: core.Sequence, primary_password: bytes, global_salt: bytes) -> tuple[bytes, str]:
    """Decrypt the provided ``core.Sequence`` with the provided Firefox primary password and salt.

    At this stage we are not yet sure of the structure of ``decoded_item``. The structure will depend on the
    ``core.Sequence`` object identifier at ``decoded_item[0][0]``, hence we extract it. This function will
    then call the apropriate ``decrypt_pbes2``or ``decrypt_sha1_triple_des_cbc`` functions to decrypt the item.

    Args:
        decoded_item: ``core.Sequence`` is a ``list`` representation of ``SEQUENCE`` as described below.
        primary_password: ``bytes`` of Firefox primary password to decrypt ciphertext with.
        global_salt: ``bytes`` of salt to prepend to primary password when calculating AES key.

    Raises:
        ValueError: When missing ``pycryptodome`` or ``asn1crypto`` dependencies.

    Returns:
        Tuple of decrypted bytes and a string representation of the identified encryption algorithm.
    """

    # SEQUENCE {
    #     SEQUENCE {
    #         OBJECTIDENTIFIER ???
    #         ...
    #     }
    #     ...
    # }

    if not HAS_CRYPTO:
        raise ValueError("Missing pycryptodome dependency")

    if not HAS_ASN1:
        raise ValueError("Missing asn1crypto depdendency")

    object_identifier = decoded_item[0][0]
    algorithm = object_identifier.dotted

    if algos.EncryptionAlgorithmId.map(algorithm) == "pbes2":
        return decrypt_pbes2(decoded_item, primary_password, global_salt), algorithm
    if algorithm == pbeWithSha1AndTripleDES_CBC:
        return decrypt_sha1_triple_des_cbc(decoded_item, primary_password, global_salt), algorithm
    # Firefox supports other algorithms (i.e. Firefox before 2018), but decrypting these is not (yet) supported.
    return b"", algorithm


def query_global_salt(key4_file: TargetPath) -> tuple[str, str]:
    with key4_file.open("rb") as fh:
        db = sqlite3.SQLite3(fh)
        for row in db.table("metadata").rows():
            if row.get("id") == "password":
                return row.get("item1", ""), row.get("item2", "")
        return None


def query_master_key(key4_file: TargetPath) -> tuple[str, str]:
    with key4_file.open("rb") as fh:
        db = sqlite3.SQLite3(fh)
        if row := next(db.table("nssPrivate").rows(), None):
            return row.get("a11", ""), row.get("a102", "")
        return None


def retrieve_master_key(primary_password: bytes, key4_file: TargetPath) -> tuple[bytes, str]:
    if not HAS_CRYPTO:
        raise ValueError("Missing pycryptodome dependency")

    if not HAS_ASN1:
        raise ValueError("Missing asn1crypto dependency")

    global_salt, password_check = query_global_salt(key4_file)
    decoded_password_check = core.load(password_check)

    try:
        decrypted_password_check, algorithm = decrypt_master_key(decoded_password_check, primary_password, global_salt)
    except EOFError:
        raise ValueError("No primary password provided")

    if not decrypted_password_check:
        raise ValueError(f"Encountered unknown algorithm {algorithm} while decrypting master key")

    expected_password_check = b"password-check\x02\x02"
    if decrypted_password_check != b"password-check\x02\x02":
        log.debug("Expected %s but got %s", expected_password_check, decrypted_password_check)
        raise ValueError("Master key decryption failed. Provided password could be missing or incorrect")

    master_key, master_key_cka = query_master_key(key4_file)
    if master_key == b"":
        raise ValueError("Password master key is not defined")

    if master_key_cka != CKA_ID:
        raise ValueError(f"Password master key CKA_ID '{master_key_cka}' is not equal to expected value '{CKA_ID}'")

    decoded_master_key = core.load(master_key)
    decrypted, algorithm = decrypt_master_key(decoded_master_key, primary_password, global_salt)
    return decrypted[:24], algorithm


def decrypt_field(key: bytes, field: tuple[bytes, bytes, bytes]) -> bytes:
    if not HAS_CRYPTO:
        raise ValueError("Missing pycryptodome dependency")

    cka, iv, ciphertext = field

    if cka != CKA_ID:
        raise ValueError(f"Expected cka to equal '{CKA_ID}' but got '{cka}'")

    return unpad(DES3.new(key, DES3.MODE_CBC, iv).decrypt(ciphertext), 8)


def decrypt(
    username: str, password: str, key4_file: TargetPath, primary_password: str = ""
) -> tuple[str | None, str | None]:
    """Decrypt a stored username and password using provided credentials and key4 file.

    Args:
        username: Encoded and encrypted password.
        password Encoded and encrypted password.
        key4_file: Path to key4.db file.
        primary_password: Password to use for decryption routine.

    Returns:
        A tuple of decoded username and password strings.

    References:
        - https://github.com/lclevy/firepwd
    """
    if not HAS_CRYPTO:
        raise ValueError("Missing pycryptodome dependency")

    if not HAS_ASN1:
        raise ValueError("Missing asn1crypto dependency")

    try:
        username = decode_login_data(username)
        password = decode_login_data(password)

        primary_password_bytes = primary_password.encode()
        key, algorithm = retrieve_master_key(primary_password_bytes, key4_file)

        if algorithm == pbeWithSha1AndTripleDES_CBC or algos.EncryptionAlgorithmId.map(algorithm) == "pbes2":
            username = decrypt_field(key, username)
            password = decrypt_field(key, password)
            return username.decode(), password.decode()

    except ValueError as e:
        raise ValueError(f"Failed to decrypt password using keyfile: {key4_file}, password: {primary_password}") from e



from __future__ import annotations

import argparse
import sys
import time
from dataclasses import dataclass
from struct import pack
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest

from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Iterator


class MqttMock(MagicMock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.disks: list = []
        self.hostname: str = ""

    def fill_disks(self, sizes: list[int]) -> None:
        self.disks = []
        pattern = list(range(8))
        for size in sizes:
            self.disks.append(bytearray(pattern) * 64 * size)  # sizes in sectors of 512

    def publish(self, topic: str, *args) -> None:
        response = MagicMock()
        tokens = topic.split("/")
        command = tokens[2]
        if command == "TOPO":
            tokens[2] = "ID"
            response.topic = "/".join(tokens)
            for host in self.hostnames:
                response.payload = host.encode("utf-8")
                self.on_message(self, None, response)
        elif tokens[2] == "INFO":
            tokens[2] = "DISKS"
            response.topic = "/".join(tokens)
            response.payload = pack("<B", len(self.disks))
            for disk in self.disks:
                response.payload += pack("<IQ", 512, len(disk))
        elif tokens[2] == "SEEK":
            tokens[2] = "READ"
            response.topic = "/".join(tokens)
            begin = int(tokens[4], 16)
            end = int(tokens[5], 16)
            response.payload = self.disks[int(tokens[3])][begin : begin + end]
        else:
            return
        self.on_message(self, None, response)


@pytest.fixture
def mock_paho(monkeypatch: pytest.MonkeyPatch) -> Iterator[MagicMock]:
    with monkeypatch.context() as m:
        if "dissect.target.loaders.mqtt" in sys.modules:
            m.delitem(sys.modules, "dissect.target.loaders.mqtt")

        mock_paho = MagicMock()
        m.setitem(sys.modules, "paho", mock_paho)
        m.setitem(sys.modules, "paho.mqtt", mock_paho.mqtt)
        m.setitem(sys.modules, "paho.mqtt.client", mock_paho.mqtt.client)

        yield mock_paho


@dataclass
class MockSeekMessage:
    data: bytes = b""


class MockBroker(MagicMock):
    _seek = False

    def seek(self, *args) -> None:
        self._seek = True

    def read(self, *args) -> MockSeekMessage | None:
        if self._seek:
            self._seek = False
            return MockSeekMessage(data=b"010101")
        return None


@pytest.fixture
def mock_broker() -> MockBroker:
    return MockBroker()


@pytest.mark.parametrize(
    ("alias", "hosts", "disks", "disk", "seek", "read", "expected"),
    [
        ("host1", ["host1"], [3], 0, 0, 3, b"\x00\x01\x02"),  # basic
        ("host2", ["host2"], [10], 0, 1, 3, b"\x01\x02\x03"),  # + use offset
        ("group1", ["host3"], [10], 0, 1, 3, b"\x01\x02\x03"),  # + use alias
        ("group2", ["host4"], [10, 10, 1], 1, 1, 3, b"\x01\x02\x03"),  # + use disk 2
        ("group3", ["host4", "host5"], [10, 10, 1], 1, 1, 3, b"\x01\x02\x03"),  # + use disk 2
    ],
)
@patch.object(time, "sleep")  # improve speed during test, no need to wait for peers
def test_remote_loader_stream(
    time: MagicMock,
    mock_paho: MagicMock,
    alias: str,
    hosts: list[str],
    disks: list[int],
    disk: int,
    seek: int,
    read: int,
    expected: bytes,
) -> None:
    mock_paho.mqtt.client.Client.return_value = MqttMock()

    from dissect.target.loaders.mqtt import Broker, MqttLoader

    broker = Broker("0.0.0.0", "1884", "key", "crt", "ca", "case1", "user", "pass")
    broker.connect()
    broker.mqtt_client.fill_disks(disks)
    broker.mqtt_client.hostnames = hosts

    with (
        patch.dict("dissect.target.loader.LOADERS_BY_SCHEME", {"mqtt": MqttLoader}),
        patch.object(MqttLoader, "broker", broker),
    ):
        targets = list(
            Target.open_all(
                [f"mqtt://{alias}?broker=0.0.0.0&port=1884&key=key&crt=crt&ca=ca&peers=1&case=case1"],
            )
        )
        assert len(targets) == len(hosts)
        target = targets[-1]
        target.disks[disk].seek(seek)
        data = target.disks[disk].read(read)
        assert data == expected


def test_mqtt_loader_prefetch(mock_broker: MockBroker, mock_paho: MagicMock) -> None:
    from dissect.target.loaders.mqtt import MqttConnection

    connection = MqttConnection(mock_broker, "")
    connection.prefetch_factor_inc = 10
    assert connection.factor == 1
    assert connection.prev == -1
    connection.read(1, 0, 100, 0)
    assert connection.factor == 1
    assert connection.prev == 0
    connection.read(1, 100, 100, 0)
    assert connection.factor == connection.prefetch_factor_inc + 1
    assert connection.prev == 100
    connection.read(1, 1200, 100, 0)
    assert connection.factor == (connection.prefetch_factor_inc * 2) + 1
    assert connection.prev == 1200


@pytest.mark.parametrize(
    ("case_name", "parse_result"),
    [
        ("valid_case_name", "valid_case_name"),
        ("ValidCase123", "ValidCase123"),
        ("valid_case_123", "valid_case_123"),
        ("invalid-case", pytest.raises(argparse.ArgumentTypeError)),
        ("invalid case", pytest.raises(argparse.ArgumentTypeError)),
        ("invalid.case", pytest.raises(argparse.ArgumentTypeError)),
        ("invalid@case", pytest.raises(argparse.ArgumentTypeError)),
        ("", pytest.raises(argparse.ArgumentTypeError)),
    ],
    ids=[
        "valid_case_lower",
        "valid_case_mixed",
        "valid_case_with_numbers",
        "invalid_case_hyphen",
        "invalid_case_space",
        "invalid_case_dot",
        "invalid_case_special_char",
        "invalid_case_empty",
    ],
)
def test_case(
    case_name: str, parse_result: str | pytest.RaisesContext[argparse.ArgumentTypeError], mock_paho: MagicMock
) -> None:
    from dissect.target.loaders.mqtt import case

    if isinstance(parse_result, str):
        assert case(case_name) == parse_result
    else:
        with parse_result:
            case(case_name)



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.log.utmp import UtmpPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_utmp_ipv6(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/os/unix/log/btmp/btmp-ipv6")
    fs_linux.map_file("var/log/btmp", data_file)

    target_linux.add_plugin(UtmpPlugin)

    results = list(target_linux.btmp())
    assert len(results) == 6

    # IPv4 address
    assert results[0].ut_host == "127.0.0.1"
    assert results[0].ut_addr == "127.0.0.1"

    # IPv6 address
    assert results[2].ut_host == "1337::1"
    assert results[2].ut_addr == "1337::1"

    # IPv6 address with 12 bytes of trailing zeroes
    assert results[5].ut_host == "1337:1::"
    assert results[5].ut_addr == "1337:1::"


def test_wtmp_plugin(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/os/unix/log/wtmp/wtmp")
    fs_linux.map_file("var/log/wtmp", data_file)

    target_linux.add_plugin(UtmpPlugin)

    results = list(target_linux.wtmp())
    assert len(results) == 70

    result = results[-1]
    assert result.ts == datetime(2021, 11, 12, 10, 12, 54, tzinfo=timezone.utc)
    assert result.ut_type == "USER_PROCESS"
    assert result.ut_user == "traut"
    assert result.ut_pid == 2176
    assert result.ut_line == ":0"
    assert result.ut_id == ""
    assert result.ut_host == ":0"
    assert result.ut_addr == "0.0.0.0"


def test_btmp_plugin(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/os/unix/log/btmp/btmp")
    fs_linux.map_file("var/log/btmp", data_file)

    target_linux.add_plugin(UtmpPlugin)

    results = list(target_linux.btmp())
    assert len(results) == 10

    result = results[-1]
    assert result.ts == datetime(2021, 11, 30, 23, 2, 9, tzinfo=timezone.utc)
    assert result.ut_type == "LOGIN_PROCESS"
    assert result.ut_user == "root"
    assert result.ut_pid == 1865007
    assert result.ut_line == "ssh:notty"
    assert result.ut_id == ""
    assert result.ut_host == "8.210.13.5"
    assert result.ut_addr == "8.210.13.5"


def test_utmp_plugin(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    """Test if we correctly parse a /var/run/utmp file."""
    fs_linux.map_file("var/run/utmp", absolute_path("_data/plugins/os/unix/log/wtmp/wtmp"))
    target_linux.add_plugin(UtmpPlugin)
    results = list(target_linux.utmp())
    assert len(results) == 70



from __future__ import annotations

import stat
from typing import TYPE_CHECKING, BinaryIO

from dissect.squashfs import INode, SquashFS, c_squashfs, exceptions

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator


class SquashFSFilesystem(Filesystem):
    __type__ = "squashfs"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.squashfs = SquashFS(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return int.from_bytes(fh.read(4), "little") == c_squashfs.c_squashfs.SQUASHFS_MAGIC

    def get(self, path: str) -> FilesystemEntry:
        return SquashFSFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: INode | None = None) -> INode:
        try:
            return self.squashfs.get(path, node)
        except exceptions.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except exceptions.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except exceptions.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except exceptions.Error as e:
            raise FileNotFoundError(path) from e


class SquashFSFilesystemEntry(FilesystemEntry):
    def get(self, path: str) -> FilesystemEntry:
        entry_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        entry = self.fs._get_node(path, self.entry)
        return SquashFSFilesystemEntry(self.fs, entry_path, entry)

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def _iterdir(self) -> Iterator[INode]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        if self.is_symlink():
            for entry in self.readlink_ext().iterdir():
                yield entry
        else:
            for entry in self.entry.iterdir():
                if entry.name in (".", ".."):
                    continue

                yield entry

    def iterdir(self) -> Iterator[str]:
        for entry in self._iterdir():
            yield entry.name

    def scandir(self) -> Iterator[FilesystemEntry]:
        for entry in self._iterdir():
            entry_path = fsutil.join(self.path, entry.name, alt_separator=self.fs.alt_separator)
            yield SquashFSFilesystemEntry(self.fs, entry_path, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.type == stat.S_IFDIR
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.type == stat.S_IFREG
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        return self.entry.type == stat.S_IFLNK

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError

        return self.entry.link

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        node = self.entry

        return fsutil.stat_result(
            [
                node.mode,
                node.inode_number,
                id(self.fs),
                getattr(node.header, "nlink", 1),
                node.uid,
                node.gid,
                node.size,
                0,  # atime
                node.mtime.timestamp(),
                0,  # ctime
            ]
        )



from __future__ import annotations

from datetime import datetime, timezone
from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch

from dissect.target.plugins.os.unix._os import UnixPlugin
from dissect.target.plugins.os.unix.trash import GnomeTrashPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_gnome_trash(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if GNOME Trash plugin finds all deleted files including recursively deleted folders and expunged items."""

    fs_unix.map_file_fh("etc/hostname", BytesIO(b"hostname"))
    fs_unix.map_dir("home/user/.local/share/Trash", absolute_path("_data/plugins/os/unix/trash"))
    target_unix_users.add_plugin(UnixPlugin)
    target_unix_users.add_plugin(GnomeTrashPlugin)

    # test if the plugin and its alias were registered
    assert target_unix_users.has_function("trash")
    assert target_unix_users.has_function("recyclebin")

    results = sorted(target_unix_users.trash(), key=lambda r: r.source)
    assert len(results) == 11

    # test if we find a deleted file
    assert results[2].ts == datetime(2024, 12, 31, 13, 37, 0, tzinfo=timezone.utc)
    assert results[2].path == "/home/user/Documents/some-location/another-file.bin"
    assert results[2].filesize == 369
    assert results[2].deleted_path == "/home/user/.local/share/Trash/files/another-file.bin"
    assert results[2].source == "/home/user/.local/share/Trash/info/another-file.bin.trashinfo"
    assert results[2].username == "user"
    assert results[2].hostname == "hostname"

    # test if we still find a file by just the .trashinfo file and no entry in the $Trash/files folder
    assert results[6].path == "/home/user/Downloads/missing-file.txt"
    assert results[6].filesize == 0
    assert results[6].source == "/home/user/.local/share/Trash/info/missing-file.txt.trashinfo"

    # test if we find a deleted directory
    assert results[7].ts == datetime(2024, 12, 31, 1, 2, 3, tzinfo=timezone.utc)
    assert results[7].path == "/home/user/Downloads/some-dir"
    assert results[7].filesize is None
    assert results[7].deleted_path == "/home/user/.local/share/Trash/files/some-dir"
    assert results[7].source == "/home/user/.local/share/Trash/info/some-dir.trashinfo"
    assert results[7].username == "user"
    assert results[7].hostname == "hostname"

    # test if we find files nested inside a deleted directory
    deleted_paths = [r.deleted_path for r in results]
    assert deleted_paths == [
        "/home/user/.local/share/Trash/expunged/123456789/some-dir",
        "/home/user/.local/share/Trash/expunged/123456789/some-dir/some-file.txt",
        "/home/user/.local/share/Trash/files/another-file.bin",
        "/home/user/.local/share/Trash/files/example.jpg",
        "/home/user/.local/share/Trash/files/file.txt.2",
        "/home/user/.local/share/Trash/files/file.txt",
        "/home/user/.local/share/Trash/files/missing-file.txt",
        "/home/user/.local/share/Trash/files/some-dir",
        "/home/user/.local/share/Trash/files/some-dir/another-dir",
        "/home/user/.local/share/Trash/files/some-dir/some-file.txt",
        "/home/user/.local/share/Trash/files/some-dir/another-dir/another-file.txt",
    ]

    # test if we find two deleted files that had the same basename
    assert results[4].path == "/home/user/Desktop/file.txt"
    assert results[4].deleted_path == "/home/user/.local/share/Trash/files/file.txt.2"
    assert results[4].filesize == 10
    assert results[5].path == "/home/user/Documents/file.txt"
    assert results[5].deleted_path == "/home/user/.local/share/Trash/files/file.txt"
    assert results[5].filesize == 20

    # test if we find expunged files
    assert results[0].path is None
    assert results[0].deleted_path == "/home/user/.local/share/Trash/expunged/123456789/some-dir"
    assert results[1].path is None
    assert results[1].deleted_path == "/home/user/.local/share/Trash/expunged/123456789/some-dir/some-file.txt"
    assert results[1].filesize == 79


def test_gnome_trash_mounts(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if GNOME Trash plugin finds Trash files in mounted devices from ``/etc/fstab``, ``/mnt`` and ``/media``."""

    fs_unix.map_file_fh("etc/hostname", BytesIO(b"hostname"))
    fs_unix.map_dir("home/user/.local/share/Trash", absolute_path("_data/plugins/os/unix/trash"))
    fs_unix.map_dir("mnt/example/.Trash-1234", absolute_path("_data/plugins/os/unix/trash"))
    fs_unix.map_dir("tmp/example/.Trash-5678", absolute_path("_data/plugins/os/unix/trash"))
    fs_unix.map_dir("media/user/example/.Trash-1000", absolute_path("_data/plugins/os/unix/trash"))

    with patch.object(target_unix_users.fs, "mounts", {"/tmp/example": None, "/mnt/example": None}):
        target_unix_users.add_plugin(UnixPlugin)
        plugin = GnomeTrashPlugin(target_unix_users)

        assert sorted([str(t) for _, t in plugin.trashes]) == [
            "/home/user/.local/share/Trash",
            "/media/user/example/.Trash-1000",
            "/mnt/example/.Trash-1234",
            "/tmp/example/.Trash-5678",
        ]

        assert len(list(plugin.trash())) == 11 * len(plugin.trashes)



"""Filesystem path manipulation functions.

Similar to posixpath and ntpath, but with support for alternative separators.
"""

from __future__ import annotations

import posixpath
import re

re_normalize_path = re.compile(r"[/]+")
re_normalize_sbs_path = re.compile(r"[\\/]+")


def normalize(path: str, alt_separator: str = "") -> str:
    if alt_separator == "\\":
        return re_normalize_sbs_path.sub("/", path)
    return re_normalize_path.sub("/", path)


def isabs(path: str, alt_separator: str = "") -> bool:
    return posixpath.isabs(normalize(path, alt_separator=alt_separator))


def join(*args, alt_separator: str = "") -> str:
    return posixpath.join(*[normalize(part, alt_separator=alt_separator) for part in args])


def split(path: str, alt_separator: str = "") -> str:
    return posixpath.split(normalize(path, alt_separator=alt_separator))


splitext = posixpath.splitext


splitdrive = posixpath.splitdrive


def splitroot(path: str, alt_separator: str = "") -> tuple[str, str, str]:
    return posixpath.splitroot(normalize(path, alt_separator=alt_separator))


def basename(path: str, alt_separator: str = "") -> str:
    return posixpath.basename(normalize(path, alt_separator=alt_separator))


def dirname(path: str, alt_separator: str = "") -> str:
    return posixpath.dirname(normalize(path, alt_separator=alt_separator))


def normpath(path: str, alt_separator: str = "") -> str:
    return posixpath.normpath(normalize(path, alt_separator=alt_separator))


def abspath(path: str, cwd: str = "", alt_separator: str = "") -> str:
    cwd = cwd or "/"
    cwd = normalize(cwd, alt_separator=alt_separator)
    path = normalize(path, alt_separator=alt_separator)
    if not isabs(path):
        path = join(cwd, path)
    return posixpath.normpath(path)


def relpath(path: str, start: str, alt_separator: str = "") -> str:
    return posixpath.relpath(
        normalize(path, alt_separator=alt_separator),
        normalize(start, alt_separator=alt_separator),
    )


def commonpath(paths: list[str], alt_separator: str = "") -> str:
    return posixpath.commonpath([normalize(path, alt_separator=alt_separator) for path in paths])


def isreserved(path: str) -> bool:
    """Return True if the path is a reserved name.

    We currently do not have any reserved names.
    """
    return False



from __future__ import annotations

import textwrap
from io import BytesIO
from typing import TYPE_CHECKING

from flow.record.fieldtypes import datetime

from dissect.target.plugins.os.unix.linux.iptables import IptablesSavePlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_iptables_plugin(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    ipv4_rules_path = "/etc/iptables/rules.v4"
    ipv4_rules = """# Generated by iptables-save v1.8.7 on Wed May 31 13:37:00 2023
    *filter
    :INPUT ACCEPT [0:0]
    :FORWARD DROP [0:0]
    :OUTPUT ACCEPT [0:0]
    :DOCKER - [0:0]
    :DOCKER-ISOLATION-STAGE-1 - [0:0]
    :DOCKER-ISOLATION-STAGE-2 - [0:0]
    :DOCKER-USER - [0:0]
    -A FORWARD -j DOCKER-USER
    -A DOCKER-ISOLATION-STAGE-1 -i br-aaaaaaaaaaaa ! -o br-bbbbbbbbbbbb -j DOCKER-ISOLATION-STAGE-2
    COMMIT
    # Completed on Wed May 31 13:37:00 2023
    # Generated by iptables-save v1.8.7 on Wed May 31 13:37:00 2023
    *nat
    :PREROUTING ACCEPT [0:0]
    :INPUT ACCEPT [0:0]
    :OUTPUT ACCEPT [0:0]
    :POSTROUTING ACCEPT [0:0]
    :DOCKER - [0:0]
    -A PREROUTING -m addrtype --dst-type LOCAL -j DOCKER
    -A DOCKER -d 127.0.0.1/32 ! -i br-aaaaaaaaaaaa -p tcp -m tcp --dport 3456 -j DNAT --to-destination 172.18.0.2:3456
    COMMIT
    # Completed on Wed May 31 13:37:00 2023
    """

    ipv6_rules_path = "/etc/iptables/rules.v6"
    ipv6_rules = """# Generated by ip6tables-save v1.8.7 on Wed Jun 1 13:37:00 2023
    *filter
    :INPUT ACCEPT [21:2975]
    :FORWARD ACCEPT [0:0]
    :OUTPUT ACCEPT [0:0]
    [78:2323] -A PREROUTING -m addrtype --dst-type LOCAL -j DOCKER
    COMMIT
    # Completed on Wed Jun 1 13:37:00 2023
    """

    fs_unix.map_file_fh(
        ipv4_rules_path,
        BytesIO(textwrap.dedent(ipv4_rules).encode()),
    )

    fs_unix.map_file_fh(
        ipv6_rules_path,
        BytesIO(textwrap.dedent(ipv6_rules).encode()),
    )

    target_unix_users.add_plugin(IptablesSavePlugin)
    results = list(target_unix_users.iptables())

    assert len(results) == 20

    assert results[0].ts == datetime(2023, 5, 31, 13, 37, 0)
    assert results[0].program == "iptables-save"
    assert results[0].version == "1.8.7"
    assert results[0].table == "filter"
    assert results[0].chain == "INPUT"
    assert results[0].type == "policy"
    assert results[0].rule == "ACCEPT"
    assert results[0].packet_count == 0
    assert results[0].byte_count == 0

    assert results[5].ts == datetime(2023, 5, 31, 13, 37, 0)
    assert results[5].program == "iptables-save"
    assert results[5].version == "1.8.7"
    assert results[5].table == "filter"
    assert results[5].chain == "DOCKER-ISOLATION-STAGE-2"
    assert results[5].type == "policy"
    assert results[5].rule == "-"
    assert results[5].packet_count == 0
    assert results[5].byte_count == 0

    assert results[15].ts == datetime(2023, 5, 31, 13, 37, 0)
    assert results[15].program == "iptables-save"
    assert results[15].version == "1.8.7"
    assert results[15].table == "nat"
    assert results[15].chain == "DOCKER"
    assert results[15].type == "rule"
    assert results[15].rule == (
        "-A DOCKER -d 127.0.0.1/32 ! -i br-aaaaaaaaaaaa -p tcp -m tcp"
        " --dport 3456 -j DNAT --to-destination 172.18.0.2:3456"
    )
    assert results[15].packet_count is None
    assert results[15].byte_count is None

    assert results[16].ts == datetime(2023, 6, 1, 13, 37, 0)
    assert results[16].program == "ip6tables-save"
    assert results[16].version == "1.8.7"
    assert results[16].table == "filter"
    assert results[16].chain == "INPUT"
    assert results[16].type == "policy"
    assert results[16].rule == "ACCEPT"
    assert results[16].packet_count == 21
    assert results[16].byte_count == 2975

    assert results[19].ts == datetime(2023, 6, 1, 13, 37, 0)
    assert results[19].program == "ip6tables-save"
    assert results[19].version == "1.8.7"
    assert results[19].table == "filter"
    assert results[19].chain == "PREROUTING"
    assert results[19].type == "rule"
    assert results[19].rule == "-A PREROUTING -m addrtype --dst-type LOCAL -j DOCKER"
    assert results[19].packet_count == 78
    assert results[19].byte_count == 2323


def test_iptables_plugin_ufw(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    ufw_rules_path = "/etc/ufw/user.rules"

    ufw_rules = """*filter
    :ufw-user-input - [0:0]
    :ufw-user-output - [0:0]
    :ufw-user-forward - [0:0]
    :ufw-user-limit - [0:0]
    :ufw-user-limit-accept - [0:0]
    ### RULES ###
    -A ufw-user-limit -m limit --limit 3/minute -j LOG --log-prefix "[UFW LIMIT BLOCK] "
    -A ufw-user-limit -j REJECT
    -A ufw-user-limit-accept -j ACCEPT
    COMMIT
    """

    fs_unix.map_file_fh(
        ufw_rules_path,
        BytesIO(textwrap.dedent(ufw_rules).encode()),
    )

    target_unix_users.add_plugin(IptablesSavePlugin)
    results = list(target_unix_users.iptables())

    # 5 policies, 3 rules
    assert len(results) == 5 + 3

    assert results[-1].source == ufw_rules_path
    assert results[-1].rule == "-A ufw-user-limit-accept -j ACCEPT"
    assert results[-1].table == "filter"
    assert results[-1].chain == "ufw-user-limit-accept"
    assert results[-1].type == "rule"



from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING

from dissect.target.filesystem import LayerFilesystem, VirtualFilesystem
from dissect.target.filesystems.dir import DirectoryFilesystem

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

log = logging.getLogger(__name__)


class Overlay2Filesystem(LayerFilesystem):
    """Docker Overlay 2 filesystem implementation.

    Deleted files will be present on the reconstructed filesystem.
    Volumes and bind mounts will be added to their respective mount locations.
    Does not support tmpfs mounts.

    References:
        - https://docs.docker.com/engine/storage/drivers/overlayfs-driver/
        - https://docs.docker.com/storage/storagedriver/
        - https://docs.docker.com/storage/volumes/
        - https://www.didactic-security.com/resources/docker-forensics.pdf
        - https://www.didactic-security.com/resources/docker-forensics-cheatsheet.pdf
        - https://github.com/google/docker-explorer
    """

    __type__ = "overlay2"

    def __init__(self, path: Path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.base_path = path

        # base_path is /foo/bar/image/overlay2/layerdb/mounts/<id> so we traverse up to /foo/bar to get to the root.
        root = path.parents[4]

        layers = []
        parent_layer = path.joinpath("parent").read_text()

        # iterate over all image layers
        while parent_layer:
            hash_type, layer_hash = parent_layer.split(":")
            layer_ref = root.joinpath("image", "overlay2", "layerdb", hash_type, layer_hash)
            cache_id = layer_ref.joinpath("cache-id").read_text()
            layers.append(("/", root.joinpath("overlay2", cache_id, "diff")))

            parent_layer = parent_file.read_text() if (parent_file := layer_ref.joinpath("parent")).exists() else None

        # add the container layers
        for container_layer_name in ["init-id", "mount-id"]:
            layer = path.joinpath(container_layer_name).read_text()
            layers.append(("/", root.joinpath("overlay2", layer, "diff")))

        # add anonymous volumes, named volumes and bind mounts
        if (config_path := root.joinpath("containers", path.name, "config.v2.json")).exists():
            try:
                config = json.loads(config_path.read_text())
            except json.JSONDecodeError as e:
                log.warning("Unable to parse overlay mounts for container %s", path.name)
                log.debug("", exc_info=e)
                return

            for mount in config.get("MountPoints").values():
                if mount["Type"] not in ["volume", "bind"]:
                    log.warning("Encountered unsupported mount type %s in container %s", mount["Type"], path.name)
                    continue

                if not mount["Source"] and mount["Name"]:
                    # anonymous volumes do not have a Source but a volume id
                    layer = root.joinpath("volumes", mount["Name"], "_data")
                elif mount["Source"]:
                    # named volumes and bind mounts have a Source set
                    layer = root.parents[-1].joinpath(mount["Source"])
                else:
                    log.warning("Could not determine layer source for mount in container %s", path.name)
                    log.debug(json.dumps(mount))
                    continue

                layers.append((mount["Destination"], layer))

        # add hosts, hostname and resolv.conf files
        for file in ["HostnamePath", "HostsPath", "ResolvConfPath"]:
            if not config.get(file) or not (fp := path.parents[-1].joinpath(config.get(file))).exists():
                log.warning("Container %s has no %s mount", path.name, file)
                continue

            layers.append(("/etc/" + fp.name, fp))

        # append and mount every layer
        for dest, layer in layers:
            # we could have collected a layer reference that actually does not exist on the host
            if not layer.exists():
                log.warning(
                    "Can not mount layer %s for container %s as it does not exist on the host", layer, path.name
                )
                continue

            # mount points can be files
            if layer.is_file():
                layer_fs = VirtualFilesystem()
                layer_fs.map_file_fh(dest, layer.open("rb"))

            # regular overlay2 layers are directories
            # mount points can be directories too
            else:
                layer_fs = DirectoryFilesystem(layer)

            log.debug("Adding layer %s to destination %s", layer, dest)
            self.append_layer().mount("/" if layer.is_file() else dest, layer_fs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.base_path}>"


class OverlayFilesystem(LayerFilesystem):
    """Podman overlay filesystem implementation.

    Currently does not support mounting of (anonymous) volumes, named volumes and bind mounts.
    Also does not map mount point files, hosts, hostname and resolv.conf files.

    References:
        - https://github.com/containers/podman
        - https://docs.podman.io/en/latest/
    """

    __type__ = "overlay"

    def __init__(self, path: Path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.base_path = path

        for dest, layer in oci_layers(path):
            if not layer.exists():
                log.warning(
                    "Can not mount layer %s for container %s as it does not exist on the host", layer, path.name
                )
                continue

            layer_fs = DirectoryFilesystem(layer)
            log.debug("Adding layer %s to destination %s", layer, dest)
            self.append_layer().mount("/" if layer.is_file() else dest, layer_fs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.base_path}>"


def oci_layers(path: Path) -> Iterator[tuple[str, Path]]:
    """Yield the layers of an OCI container provided the ``mount_path``."""
    yield ("/", path.joinpath("diff"))
    for symlink in path.joinpath("lower").read_text().split(":"):
        if symlink:
            yield ("/", path.parent.joinpath(symlink).resolve())



from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING, Callable
from unittest.mock import patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.utm import UtmLoader
from dissect.target.target import Target
from tests._utils import mkdirs

if TYPE_CHECKING:
    from pathlib import Path


@pytest.fixture
def mock_utm_dir(tmp_path: Path) -> Path:
    dummy_plist = """<?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
    <plist version="1.0">
    <dict>
        <key>Drive</key>
        <array>
            <dict>
                <key>Identifier</key>
                <string>Test</string>
                <key>ImageName</key>
                <string>Test.qcow2</string>
                <key>ImageType</key>
                <string>Disk</string>
                <key>Interface</key>
                <string>IDE</string>
                <key>InterfaceVersion</key>
                <integer>1</integer>
                <key>ReadOnly</key>
                <false/>
            </dict>
        </array>
    </dict>
    </plist>
    """

    root = tmp_path
    mkdirs(root, ["Test.utm"])
    (root / "Test.utm" / "config.plist").write_text(textwrap.dedent(dummy_plist))

    return root / "Test.utm"


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], mock_utm_dir: Path) -> None:
    """Test that we correctly use ``UtmLoader`` when opening a ``Target``."""
    with patch("dissect.target.container.open"), patch("dissect.target.target.Target.apply"):
        target = opener(mock_utm_dir)
        assert isinstance(target._loader, UtmLoader)
        assert target.path == mock_utm_dir


def test_loader(mock_utm_dir: Path) -> None:
    loader = loader_open(mock_utm_dir)
    assert isinstance(loader, UtmLoader)

    with patch("dissect.target.container.open") as mock_container_open:
        t = Target()
        loader.map(t)

        assert len(t.disks) == 1
        mock_container_open.assert_called_with(mock_utm_dir / "Data" / "Test.qcow2")



from __future__ import annotations

import re
from datetime import datetime, timezone
from functools import cached_property, lru_cache
from typing import TYPE_CHECKING

from defusedxml import ElementTree
from flow.record.base import RE_VALID_FIELD_NAME

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.helpers.fsutil import has_glob_magic
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.webserver.webserver import (
    WebserverAccessLogRecord,
    WebserverPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

LOG_RECORD_NAME = "filesystem/windows/iis/logs"

BASIC_RECORD_FIELDS = [
    ("datetime", "ts"),
    ("net.ipaddress", "client_ip"),
    ("net.ipaddress", "server_ip"),
    ("string", "username"),
    ("string", "server_name"),
    ("string", "site_name"),
    ("string", "request_method"),
    ("string", "request_path"),
    ("string", "request_query"),
    ("string", "request_size_bytes"),
    ("string", "response_size_bytes"),
    ("string", "process_time_ms"),
    # https://docs.microsoft.com/en-US/troubleshoot/iis/http-status-code
    ("string", "service_status_code"),
    # https://docs.microsoft.com/en-us/windows/win32/debug/system-error-codes?redirectedfrom=MSDN#system-error-codes-1
    ("string", "win32_status_code"),
    ("string", "format"),
    ("path", "source"),
]
BasicRecordDescriptor = TargetRecordDescriptor(LOG_RECORD_NAME, BASIC_RECORD_FIELDS)

# Simplified reverse of flow.record.base.RE_VALID_FIELD_NAME
FIELD_NAME_INVALID_CHARS_RE = re.compile(r"[^a-zA-Z0-9]")


@lru_cache(4096)
def _create_extended_descriptor(extra_fields: tuple[tuple[str, str]]) -> TargetRecordDescriptor:
    return TargetRecordDescriptor(LOG_RECORD_NAME, BASIC_RECORD_FIELDS + list(extra_fields))


class IISLogsPlugin(WebserverPlugin):
    """IIS 7 (and above) logs plugin.

    References:
        - https://docs.microsoft.com/en-us/iis/get-started/planning-your-iis-architecture/introduction-to-applicationhostconfig
        - https://docs.microsoft.com/en-us/previous-versions/iis/6.0-sdk/ms525807%28v=vs.90%29
    """

    APPLICATION_HOST_CONFIG = "%windir%/system32/inetsrv/config/applicationHost.config"

    DEFAULT_LOG_PATHS = (
        "%windir%\\System32\\LogFiles\\W3SVC*\\*.log",
        "sysvol\\Windows.old\\Windows\\System32\\LogFiles\\W3SVC*\\*.log",
        "sysvol\\inetpub\\logs\\LogFiles\\*.log",
        "sysvol\\inetpub\\logs\\LogFiles\\W3SVC*\\*.log",
        "sysvol\\Resources\\Directory\\*\\LogFiles\\Web\\W3SVC*\\*.log",
    )

    __namespace__ = "iis"

    def __init__(self, target: Target):
        super().__init__(target)
        self.config = self.target.resolve(self.APPLICATION_HOST_CONFIG)

    def check_compatible(self) -> None:
        if not self.log_dirs:
            raise UnsupportedPluginError("No IIS log files found")

    @cached_property
    def log_dirs(self) -> dict[str, set[Path]]:
        dirs = {"auto": set(), "W3C": set(), "IIS": set()}

        if (sysvol_files := self.target.fs.path("sysvol/files")).exists():
            dirs["auto"].add(sysvol_files)

        try:
            xml_data = ElementTree.fromstring(self.config.read_bytes(), forbid_dtd=True)
            for log_file_element in xml_data.findall("*/sites/*/logFile"):
                log_format = log_file_element.get("logFormat") or "W3C"
                if log_dir := log_file_element.get("directory"):
                    if log_format not in dirs:
                        self.target.log.warning("Unsupported log format %s, skipping %s", log_format, log_dir)
                        continue
                    dirs[log_format].add(self.target.resolve(log_dir))

        except (ElementTree.ParseError, FileNotFoundError) as e:
            self.target.log.warning("Error while parsing %s", self.config)
            self.target.log.debug("", exc_info=e)

        for log_path in self.DEFAULT_LOG_PATHS:
            try:
                # later on we use */*.log to collect the files, so we need to move up 2 levels
                log_path = self.target.expand_env(log_path)
                log_dir = self.target.fs.path(log_path).parents[1]
            except IndexError:
                self.target.log.info("Incompatible path found: %s", log_path)
                continue

            if not has_glob_magic(str(log_dir)) and log_dir.exists():
                dirs["auto"].add(log_dir)
                continue

            for _log_dir_str in self.target.fs.glob(str(log_dir)):
                if not (_log_dir := self.target.fs.path(_log_dir_str)).is_dir():
                    continue
                dirs["auto"].add(_log_dir)

        return dirs

    @export(record=BasicRecordDescriptor)
    def logs(self) -> Iterator[TargetRecordDescriptor]:
        """Return contents of IIS (v7 and above) log files.

        Internet Information Services (IIS) for Windows Server is a manageable web server for hosting anything on the
        web. Logs files might, for example, contain traces that indicate that the web server has been exploited.

        Supported log formats: IIS, W3C.
        """

        parsers = {
            "W3C": parse_w3c_format_log,
            "IIS": parse_iis_format_log,
            "auto": parse_autodetect_format_log,
        }

        for format in ("IIS", "W3C", "auto"):
            for log_dir in self.log_dirs.get(format, ()):
                for log_file in log_dir.glob("*/*.log"):
                    self.target.log.info("Processing IIS log file %s in %s format", log_file, format)
                    yield from parsers[format](self.target, log_file)

        # We don't implement _get_paths() in the IIS plugin because there's little use for it for the way the plugin
        # is currently implemented. So handle direct files here.
        if self.target.is_direct:
            for log_file in self.get_paths():
                yield from parse_autodetect_format_log(self.target, log_file)

    @export(record=WebserverAccessLogRecord)
    def access(self) -> Iterator[WebserverAccessLogRecord]:
        """Return contents of IIS (v7 and above) log files in unified WebserverAccessLogRecord format.

        See function ``iis.logs`` for more information and more verbose IIS records.
        """
        for iis_record in self.logs():
            yield WebserverAccessLogRecord(
                ts=iis_record.ts,
                remote_user=iis_record.username,
                remote_ip=iis_record.client_ip,
                method=iis_record.request_method,
                uri=iis_record.request_path,
                query=getattr(iis_record, "request_query", None),
                protocol=getattr(iis_record, "cs_version", None),
                status_code=getattr(iis_record, "service_status_code", None),
                bytes_sent=iis_record.response_size_bytes,
                referer=getattr(iis_record, "cs_referer", None),
                useragent=getattr(iis_record, "cs_user_agent", None),
                source=iis_record.source,
                _target=self.target,
            )


def parse_autodetect_format_log(target: Target, path: Path) -> Iterator[BasicRecordDescriptor]:
    with path.open() as fh:
        is_w3c = fh.read(1) == b"#"

    if is_w3c:
        yield from parse_w3c_format_log(target, path)
    else:
        yield from parse_iis_format_log(target, path)


def parse_w3c_format_log(target: Target, path: Path) -> Iterator[TargetRecordDescriptor]:
    """Parse log file in W3C format and yield log records.

    This is the default logging format for IIS [^3].

    References:
        - https://docs.microsoft.com/en-us/previous-versions/iis/6.0-sdk/ms525807%28v=vs.90%29#w3c-extended-log-file-format
        - https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2003/cc786596%28v=ws.10%29
        - https://learn.microsoft.com/en-us/iis/configuration/system.applicationHost/sites/site/logFile
    """

    basic_fields = {
        "c-ip",
        "s-ip",
        "cs-username",
        "s-computername",
        "s-sitename",
        "cs-method",
        "cs-uri-stem",
        "cs-uri-query",
        "cs-bytes",
        "sc-bytes",
        "time-taken",
        "sc-status",
        "sc-win32-status",
    }

    record_descriptor = None
    fields = []
    extra_fields = []
    for line in path.open().readlines():
        line = line.decode("utf-8", errors="backslashreplace").strip()

        if line.startswith("#Fields"):
            _, _, fields_str = line.partition("Fields: ")
            fields = fields_str.split()
            extra_fields = sorted(set(fields) - basic_fields)
            extra_fields_with_types = [("string", normalise_field_name(f)) for f in extra_fields]
            record_descriptor = _create_extended_descriptor(tuple(extra_fields_with_types))
            continue
        elif line.startswith("#"):
            continue

        values = line.split()

        if len(values) != len(fields):
            target.log.warning("Log values do not match fields defined, skipping: %r", line)
            continue

        if not record_descriptor:
            target.log.warning("Comment line with the fields defined should come before the values, skipping: %r", line)

        raw = replace_dash_with_none(dict(zip(fields, values)))

        # Example:
        # {
        #     "c-ip": "127.0.0.1",
        #     "cs(Cookie)": null,
        #     "cs(Referer)": null,
        #     "cs(User-Agent)": "Mozilla/5.0+(Windows+NT+10.0;+Win64;+x64)+AppleWebKit/537.36+(KHTML,+like+Gecko)+Chrome/93.0.4577.82+Safari/537.36+Edg/93.0.961.52",  # noqa: E501
        #     "cs-bytes": "714",
        #     "cs-host": "127.0.0.1",
        #     "cs-method": "GET",
        #     "cs-uri-query": null,
        #     "cs-uri-stem": "/nonexistent-path/path+path2",
        #     "cs-username": null,
        #     "cs-version": "HTTP/1.1",
        #     "custom-field-1": null,
        #     "custom-field-2": "Cache-Control:+max-age=0++Connection:+keep-alive++Accept:+text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9++Accept-Encoding:+gzip,+deflate,+br++Accept-Language:+en-US,en;q=0.9++Host:+127.0.0.1++User-Agent:+Mozilla/5.0+(Windows+NT+10.0;+Win64;+x64)+AppleWebKit/537.36+(KHTML,+like+Gecko)+Chrome/93.0.4577.82+Safari/537.36+Edg/93.0.961.52++sec-ch-ua:+\"Microsoft+Edge\";v=\"93\",+\"+Not;A+Brand\";v=\"99\",+\"Chromium\";v=\"93\"++sec-ch-ua-mobile:+?0++sec-ch-ua-platform:+\"Windows\"++Upgrade-Insecure-Requests:+1++Sec-Fetch-Site:+none++Sec-Fetch-Mode:+navigate++Sec-Fetch-User:+?1++Sec-Fetch-Dest:+document++",  # noqa: E501
        #     "date": "2021-10-01",
        #     "s-computername": "DESKTOP-PJOQLJS",
        #     "s-ip": "127.0.0.1",
        #     "s-port": "80",
        #     "s-sitename": "W3SVC1",
        #     "sc-bytes": "5143",
        #     "sc-status": "404",
        #     "sc-substatus": "11",
        #     "sc-win32-status": "0",
        #     "time": "18:03:57",
        #     "time-taken": "1"
        # }

        # Make the datetime timezone aware.
        # "the time stamp for each W3C Extended Logging log record is UTC-based." [^3]
        ts = None
        if raw.get("date") and raw.get("time"):
            ts = datetime.strptime(f"{raw['date']} {raw['time']}", "%Y-%m-%d %H:%M:%S").replace(tzinfo=timezone.utc)

        yield record_descriptor(
            ts=ts,
            client_ip=raw.get("c-ip"),
            server_ip=raw.get("s-ip"),
            username=raw.get("cs-username"),
            server_name=raw.get("s-computername"),
            site_name=raw.get("s-sitename"),
            request_method=raw.get("cs-method"),
            request_path=raw.get("cs-uri-stem"),
            request_query=raw.get("cs-uri-query"),
            request_size_bytes=raw.get("cs-bytes"),
            response_size_bytes=raw.get("sc-bytes"),
            process_time_ms=raw.get("time-taken"),
            service_status_code=raw.get("sc-status"),
            win32_status_code=raw.get("sc-win32-status"),
            format="W3C",
            source=path,
            _target=target,
            **{normalise_field_name(field): raw.get(field) for field in extra_fields},
        )


def parse_iis_format_log(target: Target, path: Path) -> Iterator[BasicRecordDescriptor]:
    """Parse log file in IIS format and stream log records.

    This format is not the default IIS log format.

    References:
        - https://docs.microsoft.com/en-us/previous-versions/iis/6.0-sdk/ms525807%28v=vs.90%29#iis-log-file-format
        - https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2003/cc728311%28v=ws.10%29
        - https://learn.microsoft.com/en-us/iis/configuration/system.applicationHost/sites/site/logFile
    """

    target_tz = target.datetime.tzinfo

    for line in path.open().readlines():
        # even though the docs say that IIS log format is ASCII format,
        # it is possible to select UTF-8 in configuration
        line = line.decode("utf-8", errors="backslashreplace")

        # Example:
        # 127.0.0.1, -, 9/20/2021, 8:51:21, W3SVC1, DESKTOP-PJOQLJS, 127.0.0.1, 0, 691, 5005, 404, 2, GET, /some, -,
        parts = [part.strip() for part in line.strip().split(",")]

        if len(parts) != 16:
            target.log.debug("Unrecognised log line format, skipping: %r", line)
            continue

        row = {
            "client_ip": parts[0],
            "username": parts[1],
            "ts": parse_iis_datetime(parts[2], parts[3], target_tz),
            "site_name": parts[4],
            "server_name": parts[5],
            "server_ip": parts[6],
            "process_time_ms": parts[7],
            "request_size_bytes": parts[8],
            "response_size_bytes": parts[9],
            "service_status_code": parts[10],
            "win32_status_code": parts[11],
            "request_method": parts[12],
            "request_path": parts[13],
            "request_query": parts[14],
            "format": "IIS",
            "source": path,
            "_target": target,
        }
        row = replace_dash_with_none(row)
        yield BasicRecordDescriptor(**row)


def parse_iis_datetime(date_str: str, time_str: str, tzinfo: timezone) -> datetime:
    # Example: 10/1/2021 7:19:59
    # "time is recorded as local time." [^3]
    return datetime.strptime(f"{date_str} {time_str}", "%m/%d/%Y %H:%M:%S").replace(tzinfo=tzinfo)


def replace_dash_with_none(data: dict) -> dict:
    """Replace ``-`` placeholder in dictionary values with ``None``."""
    return {k: (None if v == "-" else v) for k, v in data.items()}


def normalise_field_name(field: str) -> str:
    """Replace all character that are not allowed in the field name by flow.record
    with _, and strip all hanging _ from start / end of the string.
    """
    if RE_VALID_FIELD_NAME.match(field):
        return field

    return FIELD_NAME_INVALID_CHARS_RE.sub("_", field).strip("_").lower()



from __future__ import annotations

from typing import BinaryIO

from dissect.hypervisor.util import vmtar

from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.helpers.fsutil import open_decompress


class VmtarFilesystem(TarFilesystem):
    __type__ = "vmtar"

    def __init__(self, fh: BinaryIO, base: str | None = None, *args, **kwargs):
        fh = open_decompress(fileobj=open_decompress(fileobj=fh))
        super().__init__(fh, base, *args, tarinfo=vmtar.VisorTarInfo, **kwargs)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        """Detect a vmtar file on a given file-like object."""
        # vmtar files can be double compressed (gzip + lzma)
        fh = open_decompress(fileobj=open_decompress(fileobj=fh))

        fh.seek(257)
        return fh.read(8) == b"visor  \x00"



from __future__ import annotations

import gzip
import hashlib
import struct
from base64 import b64decode
from datetime import datetime, timezone
from io import BytesIO
from tarfile import ReadError
from typing import TYPE_CHECKING, BinaryIO, TextIO

from dissect.util import cpio
from dissect.util.compression import xz

from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.helpers.record import TargetRecordDescriptor, UnixUserRecord
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.plugins.os.unix.linux.fortios._keys import (
    KERNEL_KEY_MAP,
    AesKey,
    ChaCha20Key,
    ChaCha20Seed,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target

try:
    from Crypto.Cipher import AES, ChaCha20
    from Crypto.Util import Counter

    HAS_CRYPTO = True
except ImportError:
    HAS_CRYPTO = False

FortiOSUserRecord = TargetRecordDescriptor(
    "fortios/user",
    [
        ("string", "name"),
        ("string[]", "groups"),
        ("string", "password"),
        ("path", "home"),
    ],
)


def create_tar_filesystem(fileobj: BinaryIO) -> TarFilesystem:
    """Create appropriate ``TarFilesystem`` based on file format.

    Args:
        fileobj: The file-like object of a tar or cpio file

    Returns:
        TarFilesystem with cpio handler if cpio format is detected.
    """
    if open_decompress(fileobj=fileobj).read(4) == b"0707":
        return TarFilesystem(fileobj, tarinfo=cpio.CpioInfo)
    return TarFilesystem(fileobj)


class FortiOSPlugin(LinuxPlugin):
    """FortiOS plugin for various Fortinet appliances."""

    def __init__(self, target: Target):
        super().__init__(target)
        self._version = None
        self._config = self._load_config()

    def _load_config(self) -> dict:
        CONFIG_FILES = {
            "/data/system.conf": "global-config",  # FortiManager
            "/data/config/daemon.conf.gz": "daemon",  # FortiOS 4.x
            "/data/config/sys_global.conf.gz": "global-config",  # Seen in FortiOS 5.x - 7.x
            "/data/config/sys_vd_root.conf.gz": "root-config",  # FortiOS 4.x
            "/data/config/sys_vd_root+root.conf.gz": "root-config",  # Seen in FortiOS 6.x - 7.x
            "/data/config/global_system_interface.gz": "interfaces",  # Seen in FortiOS 5.x - 7.x
        }

        config = {}
        for conf_file, section in CONFIG_FILES.items():
            if (conf_path := self.target.fs.path(conf_file)).exists():
                with gzip.open(conf_path.open("rb"), "rt") if conf_file.endswith("gz") else conf_path.open("rt") as fh:
                    if not self._version and section in ["global-config", "root-config"]:
                        self._version = fh.readline().split("=", 1)[1]

                    parsed = FortiOSConfig.from_fh(fh)
                    config |= {section: parsed} if section else parsed

        return config

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            # Tested on FortiGate, FortiAnalyzer and FortiManager.
            # Other Fortinet devices may look different.
            if fs.exists("/rootfs.gz") and (any(map(fs.exists, (".fgtsum", ".fmg_sign", "flatkc", "system.conf")))):
                return fs
        return None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        target.log.warning("Attempting to load rootfs.gz, this can take a while")
        rootfs = sysvol.path("/rootfs.gz")
        vfs = None

        try:
            vfs = create_tar_filesystem(rootfs.open())
        except ReadError:
            # The rootfs.gz file could be encrypted.
            try:
                kernel_hash = get_kernel_hash(sysvol)
                target.log.info("Kernel hash: %s", kernel_hash)
                key = key_iv_for_kernel_hash(kernel_hash)
                target.log.info("Trying to decrypt_rootfs using key: %r", key)
                rfs_fh = decrypt_rootfs(rootfs.open(), key)
                target.log.info("Decrypted fh: %r", rfs_fh)
                vfs = create_tar_filesystem(rfs_fh)
            except RuntimeError:
                target.log.warning("Could not decrypt rootfs.gz, missing `pycryptodome` dependency")
            except ValueError as e:
                target.log.warning("Could not decrypt rootfs.gz, unknown kernel hash (%s)", kernel_hash)
                target.log.debug("", exc_info=e)
            except ReadError as e:
                target.log.warning("Could not mount rootfs.gz, it could be corrupt")
                target.log.debug("", exc_info=e)

        if vfs:
            target.fs.mount("/", vfs)

        target.fs.mount("/data", sysvol)

        # FortiGate
        if (datafs_tar := sysvol.path("/datafs.tar.gz")).exists():
            target.fs.append_layer().mount("/data", TarFilesystem(BytesIO(datafs_tar.open("rb").read())))

        # Additional FortiGate or FortiManager tars with corrupt XZ streams
        target.log.warning("Attempting to load XZ files, this can take a while")
        for path in (
            "bin.tar.xz",
            "usr.tar.xz",
            "migadmin.tar.xz",
            "node-scripts.tar.xz",
            "docker.tar.xz",
            "syntax.tar.xz",
        ):
            if (tar := target.fs.path(path)).exists() or (tar := sysvol.path(path)).exists():
                fh = xz.repair_checksum(BytesIO(tar.open("rb").read()))
                target.fs.append_layer().mount("/", TarFilesystem(fh))

        # FortiAnalyzer and FortiManager
        if (rootfs_ext_tar := sysvol.path("rootfs-ext.tar.xz")).exists():
            target.fs.append_layer().mount("/", TarFilesystem(rootfs_ext_tar.open("rb")))

        # If there is no /migadmin at this point, check sysvol and map accordingly
        if not target.fs.exists("/migadmin") and target.fs.exists("/data/migadmin"):
            target.log.info("Directory migadmin found in sysvol, mapping /migadmin to /data/migadmin")
            target.fs.map_file_entry("/migadmin", target.fs.get("/data/migadmin"))

        # Filesystem mounts can be discovered in the FortiCare debug report
        # or using ``fnsysctl ls`` and ``fnsysctl df`` in the cli.
        for fs in target.filesystems:
            # log partition
            if fs.__type__ == "ext" and (
                fs.extfs.volume_name.startswith("LOGUSEDX") or fs.path("/root/clog").is_symlink()
            ):
                target.fs.mount("/var/log", fs)

            # EFI partition
            if fs.__type__ == "fat" and fs.path("/EFI").exists():
                target.fs.mount("/boot", fs)

            # data2 partition
            if fs.__type__ == "ext" and (
                (fs.path("/new_alert_msg").exists() and fs.path("/template").exists())  # FortiGate
                or (fs.path("/swapfile").exists() and fs.path("/old_fmversion").exists())  # FortiManager
            ):
                target.fs.mount("/data2", fs)

        # Symlink unix-like paths
        unix_paths = [("/data/passwd", "/etc/passwd")]
        for src, dst in unix_paths:
            if target.fs.path(src).exists() and not target.fs.path(dst).exists():
                target.fs.symlink(src, dst)

        return cls(target)

    @export(property=True)
    def hostname(self) -> str | None:
        """Return configured hostname."""
        try:
            return self._config["global-config"]["system"]["global"]["hostname"][0]
        except KeyError:
            return None

    @export(property=True)
    def ips(self) -> list[str]:
        """Return IP addresses of configured interfaces."""
        result = []

        try:
            # FortiOS 6 and 7 (from global_system_interface.gz)
            for key, iface in self._config["interfaces"].items():
                if not key.startswith("port"):
                    continue
                result += [ip for ip in iface["ip"] if not ip.startswith("255")]
        except KeyError as e:
            self.target.log.debug("Exception while parsing FortiOS interfaces", exc_info=e)

        try:
            # Other versions
            for conf in self._config["global-config"]["system"]["interface"].values():
                if "ip" in conf:
                    result.append(conf.ip[0])
        except KeyError as e:
            self.target.log.debug("Exception while parsing FortiOS system interfaces", exc_info=e)

        return result

    @export(property=True)
    def dns(self) -> list[str]:
        """Return configured WAN DNS servers."""
        try:
            entries = [entry[0] for entry in self._config["global-config"]["system"]["dns"].values()]
        except KeyError:
            entries = []
        return entries

    @export(property=True)
    def version(self) -> str:
        """Return FortiOS version."""
        if self._version:
            return parse_version(self._version)
        return "FortiOS Unknown"

    @export(record=FortiOSUserRecord)
    def users(self) -> Iterator[FortiOSUserRecord | UnixUserRecord]:
        """Return local users of the FortiOS system."""

        # Possible unix-like users
        yield from super().users()

        # FortiGate administrative users
        try:
            for username, entry in self._config["global-config"]["system"]["admin"].items():
                yield FortiOSUserRecord(
                    name=username,
                    password=":".join(entry.get("password", [])),
                    groups=list(entry.get("accprofile", [])),
                    home="/root",
                    _target=self.target,
                )
        except KeyError as e:
            self.target.log.warning("Exception while parsing FortiOS admin users")
            self.target.log.debug("", exc_info=e)

        # FortiManager administrative users
        if self._config.get("global-config", {}).get("system", {}).get("admin", {}).get("user"):
            try:
                for username, entry in self._config["global-config"]["system"]["admin"]["user"].items():
                    yield FortiOSUserRecord(
                        name=username,
                        password=":".join(entry.get("password", [])),
                        groups=list(entry.get("profileid", [])),
                        home="/root",
                        _target=self.target,
                    )
            except KeyError as e:
                self.target.log.warning("Exception while parsing FortiManager admin users")
                self.target.log.debug("", exc_info=e)

        if self._config.get("root-config", {}).get("user", {}).get("local"):
            # Local users
            try:
                local_groups = local_groups_to_users(self._config["root-config"]["user"]["group"])
            except KeyError as e:
                self.target.log.warning("Unable to get local user groups in root config")
                self.target.log.debug("", exc_info=e)
                local_groups = {}

            try:
                for username, entry in self._config["root-config"]["user"].get("local", {}).items():
                    try:
                        password = decrypt_password(entry["passwd"][-1])
                    except (ValueError, RuntimeError):
                        password = ":".join(entry.get("passwd", []))

                    yield FortiOSUserRecord(
                        name=username,
                        password=password,
                        groups=local_groups.get(username, []),
                        home=None,
                        _target=self.target,
                    )
            except KeyError as e:
                self.target.log.warning("Exception while parsing FortiOS local users")
                self.target.log.debug("", exc_info=e)

        if self._config.get("root-config", {}).get("user", {}).get("group", {}).get("guestgroup"):
            # Temporary guest users
            try:
                for entry in (
                    self._config["root-config"]["user"]["group"].get("guestgroup", {}).get("guest", {}).values()
                ):
                    try:
                        password = decrypt_password(entry.get("password")[-1])
                    except (ValueError, RuntimeError):
                        password = ":".join(entry.get("password"))

                    yield FortiOSUserRecord(
                        name=entry["user-id"][0],
                        password=password,
                        groups=["guestgroup"],
                        home=None,
                        _target=self.target,
                    )
            except KeyError as e:
                self.target.log.warning("Exception while parsing FortiOS temporary guest users")
                self.target.log.debug("", exc_info=e)

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.FORTIOS.value

    @export(property=True)
    def architecture(self) -> str | None:
        """Return architecture FortiOS runs on."""
        for path in ["/lib/libav.so", "/bin/ctr", "/bin/grep"]:
            if (bin := self.target.fs.path(path)).exists():
                return self._get_architecture(path=bin)
        return None


class ConfigNode(dict):
    def set(self, path: list[str], value: str) -> None:
        node = self

        for part in path[:-1]:
            if part not in node:
                node[part] = ConfigNode()
            node = node[part]

        node[path[-1]] = value

    def __getattr__(self, attr: str) -> ConfigNode | str:
        return self[attr]


class FortiOSConfig(ConfigNode):
    @classmethod
    def from_fh(cls, fh: TextIO) -> Self:
        root = cls()

        stack = []
        for parts in _parse_config(fh):
            cmd = parts[0]

            if cmd == "config":
                if parts[1] == "vdom" and stack == [["vdom"]]:
                    continue

                stack.append(parts[1:])

            elif cmd == "edit":
                stack.append(parts[1:])

            elif cmd == "end" or cmd == "next":
                if stack:
                    stack.pop()

            elif cmd == "set":
                path = []
                for part in stack:
                    path += part

                path.append(parts[1])
                root.set(path, parts[2:])

        return root


def _parse_config(fh: TextIO) -> Iterator[list[str]]:
    parts = []
    string = None

    for line in fh:
        if not (line := line.strip()) or line.startswith("#"):
            continue

        for part in line.split(" "):
            if part.startswith('"'):
                if part.endswith('"'):
                    parts.append(part[1:-1])
                else:
                    string = [part[1:]]
            elif part.endswith('"') and part[-2] != "\\":
                string.append(part[:-1])
                parts.append(" ".join(string))
                string = None
            elif string:
                string.append(part)
            else:
                parts.append(part)

        if string:
            string.append("\n")

        if parts and not string:
            yield parts
            parts = []


def parse_version(input: str) -> str:
    """Attempt to parse the config FortiOS version to a readable format.

    The input ``FGVM64-7.4.1-FW-build2463-230830:opmode=0:vdom=0`` would
    return the following output: ``FortiGate VM 7.4.1 (build 2463, 2023-08-30)``.

    References:
        - https://support.fortinet.com/Download/VMImages.aspx
    """

    PREFIXES = {
        "FGV": "FortiGate VM",  # FGVM64
        "FGT": "FortiGate",  # can also be FGT-VM in 4.x/5.x
        "FMG": "FortiManager",
        "FAZ": "FortiAnalyzer",
        "FFW": "FortiFirewall",
        "FOS": "FortiOS",
        "FWB": "FortiWeb",
        "FAD": "FortiADC",
    }

    try:
        version_str = input.split(":", 1)[0].strip()
        type, version, _, build_num, build_date = version_str.rsplit("-", 4)

        build_num = build_num.replace("build", "build ", 1)
        build_date = datetime.strptime(build_date, "%y%m%d").replace(tzinfo=timezone.utc).strftime("%Y-%m-%d")
    except ValueError:
        return input
    else:
        type = PREFIXES.get(type[:3], type)
        return f"{type} {version} ({build_num}, {build_date})"


def local_groups_to_users(config_groups: dict) -> dict:
    """Map FortiOS groups to a dict with usernames as key."""
    user_groups = {}
    for group, items in config_groups.items():
        for user in items.get("member", []):
            if user in user_groups:
                user_groups[user].append(group)
            else:
                user_groups[user] = [group]
    return user_groups


def decrypt_password(input: str) -> str:
    """Decrypt FortiOS encrypted secrets.

    Works for FortiGate 5.x, 6.x and 7.x (CVE-2019-6693).

    NOTE:
        - FortiManager uses a 16-byte IV and is not supported (CVE-2020-9289).
        - FortiGate 4.x uses DES and a static 8-byte key and is not supported.

    Returns decoded plaintext or original input ciphertext when decryption failed.

    References:
        - https://www.fortiguard.com/psirt/FG-IR-19-007
    """

    if not HAS_CRYPTO:
        raise RuntimeError("Missing pycryptodome dependency")

    if input[:3] in ["SH2", "AK1"]:
        raise ValueError("Password is a hash (SHA-256 or SHA-1) and cannot be decrypted.")

    ciphertext = b64decode(input)
    iv = ciphertext[:4] + b"\x00" * 12
    key = b"Mary had a littl"
    cipher = AES.new(key, iv=iv, mode=AES.MODE_CBC)
    plaintext = cipher.decrypt(ciphertext[4:])

    try:
        return plaintext.split(b"\x00", 1)[0].decode()
    except UnicodeDecodeError:
        return "ENC:" + input


def key_iv_for_kernel_hash(kernel_hash: str) -> AesKey | ChaCha20Key:
    """Return decryption key and IV for a specific sha256 kernel hash.

    The decryption key and IV are used to decrypt the ``rootfs.gz`` file.

    Args:
        kernel_hash: SHA256 hash of the kernel file.

    Returns:
        Tuple with decryption key and IV.

    Raises:
        ValueError: When no decryption keys are available for the given kernel hash.
    """

    key = KERNEL_KEY_MAP.get(kernel_hash)
    if isinstance(key, ChaCha20Seed):
        # FortiOS 7.4.x uses a KDF to derive the key and IV
        key, iv = _kdf_7_4_x(key.key)
        return ChaCha20Key(key, iv)
    if isinstance(key, ChaCha20Key):
        # FortiOS 7.0.13 and 7.0.14 uses a static key and IV
        return key
    if isinstance(key, AesKey):
        # FortiOS 7.0.16, 7.2.9, 7.4.4, 7.6.0 and higher uses AES-CTR with a custom CTR increment
        return key
    raise ValueError(f"No known decryption keys for kernel hash: {kernel_hash}")


def chacha20_decrypt(fh: BinaryIO, key: ChaCha20Key) -> bytes:
    """Decrypt file using ChaCha20 with given ChaCha20Key.

    Args:
        fh: File-like object to the encrypted rootfs.gz file.
        key: ChaCha20Key.

    Returns:
        Decrypted bytes.
    """

    # First 8 bytes = counter, last 8 bytes = nonce
    # PyCryptodome interally divides this seek by 64 to get a (position, offset) tuple
    # We're interested in updating the position in the ChaCha20 internal state, so to make
    # PyCryptodome "OpenSSL-compatible" we have to multiply the counter by 64
    cipher = ChaCha20.new(key=key.key, nonce=key.iv[8:])
    cipher.seek(int.from_bytes(key.iv[:8], "little") * 64)
    return cipher.decrypt(fh.read())


def calculate_counter_increment(iv: bytes) -> int:
    """Calculate the custom FortiGate CTR increment from IV.

    Args:
        iv: 16 bytes IV.

    Returns:
        Custom CTR increment.
    """
    increment = 0
    for i in range(16):
        increment ^= (iv[i] & 15) ^ ((iv[i] >> 4) & 0xFF)
    return max(increment, 1)


def aes_decrypt(fh: BinaryIO, key: AesKey) -> bytes:
    """Decrypt file using a custom AES CTR increment with given AesKey.

    Args:
        fh: File-like object to the encrypted rootfs.gz file.
        key: AesKey.

    Returns:
        Decrypted bytes.
    """

    data = bytearray(fh.read())

    # Calculate custom CTR increment from IV
    increment = calculate_counter_increment(key.iv)
    advance_block = (b"\x69" * 16) * (increment - 1)

    # AES counter is little-endian and has a prefix
    prefix, counter = struct.unpack("<8sQ", key.iv)
    ctr = Counter.new(
        64,
        prefix=prefix,
        initial_value=counter,
        little_endian=True,
        allow_wraparound=True,
    )
    cipher = AES.new(key.key, mode=AES.MODE_CTR, counter=ctr)

    nblocks, nleft = divmod(len(data), 16)
    for i in range(nblocks):
        offset = i * 16
        data[offset : offset + 16] = cipher.decrypt(data[offset : offset + 16])
        cipher.decrypt(advance_block)  # custom advance the counter

    if nleft:
        data[nblocks * 16 :] = cipher.decrypt(data[nblocks * 16 :])

    return data


def decrypt_rootfs(fh: BinaryIO, key: ChaCha20Key | AesKey) -> BinaryIO:
    """Attempt to decrypt an encrypted ``rootfs.gz`` file with given key and IV.

    FortiOS releases as of 7.4.1 / 2023-08-31, have ChaCha20 encrypted ``rootfs.gz`` files.
    This function attempts to decrypt a ``rootfs.gz`` file using a static key and IV
    which can be found in the kernel.

    Known keys can be found in the ``_keys.py`` file.

    References:
        - https://docs.fortinet.com/document/fortimanager/7.4.2/release-notes/519207/special-notices
        - Reversing kernel (fgt_verifier_iv, fgt_verifier_decrypt, fgt_verifier_initrd)

    Args:
        fh: File-like object to the encrypted rootfs.gz file.
        key: ChaCha20Key or AesKey.

    Returns:
        File-like object to the decrypted rootfs.gz file.

    Raises:
        ValueError: When decryption failed.
        RuntimeError: When PyCryptodome is not available.
    """

    if not HAS_CRYPTO:
        raise RuntimeError("Missing pycryptodome dependency")

    result = b""
    if isinstance(key, ChaCha20Key):
        result = chacha20_decrypt(fh, key)
    elif isinstance(key, AesKey):
        result = aes_decrypt(fh, key)
        result = result[:-256]  # strip off the 256 byte footer

    if result[0:2] != b"\x1f\x8b":
        raise ValueError("Failed to decrypt: No gzip magic header found.")

    return BytesIO(result)


def _kdf_7_4_x(key_data: str | bytes, offset_key: int = 4, offset_iv: int = 5) -> tuple[bytes, bytes]:
    """Derive 32 byte key and 16 byte IV from 32 byte seed.

    As the IV needs to be 16 bytes, we return the first 16 bytes of the sha256 hash.
    """

    if isinstance(key_data, str):
        key_data = bytes.fromhex(key_data)

    key = hashlib.sha256(key_data[offset_key:32] + key_data[:offset_key]).digest()
    iv = hashlib.sha256(key_data[offset_iv:32] + key_data[:offset_iv]).digest()[:16]
    return key, iv


def get_kernel_hash(sysvol: Filesystem) -> str | None:
    """Return the SHA256 hash of the (compressed) kernel."""
    kernel_files = ["flatkc", "vmlinuz", "vmlinux"]
    for k in kernel_files:
        if sysvol.path(k).exists():
            return sysvol.sha256(k)
    return None



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct
from dissect.util.ts import from_unix

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.locate.locate import BaseLocatePlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

# References: https://linux.die.net/man/5/locate.db
mlocate_def = """
#define MAGIC 0x006d6c6f63617465             /* b'/x00mlocate' */

struct header_config {
    int32 conf_size;
    int8 version;                            /* file format version */
    int8 require_visibility;
    int8 pad0[2];                             /* 32-bit total alignment */
    char root_database;
    char config_block[conf_size];
    int8 pad1;
};

enum DBE_TYPE: uint8 {                       /* database entry type */
    FILE         = 0x0,                      /* file */
    DIRECTORY    = 0x1,                      /* directory */
    END          = 0x2                       /* end of directory */
};

struct directory_entry {
    /* time is the 'maximum of st_ctime and
       st_mtime of the directory' according to docs */
    int64 time_seconds;
    int32 time_nanoseconds;
    int32 padding;
    char path[];
};

struct entry {
    char path[];
};
"""


@dataclass
class MLocate:
    ts: datetime
    ts_ns: int
    parent: str
    path: str
    dbe_type: str


MLocateRecord = TargetRecordDescriptor(
    "linux/locate/mlocate",
    [
        ("datetime", "ts"),
        ("varint", "ts_ns"),
        ("path", "parent"),
        ("path", "path"),
        ("string", "type"),
        ("string", "source"),
    ],
)

c_mlocate = cstruct(endian=">").load(mlocate_def)


class MLocateFile:
    """Parser for mlocate files.

    References:
        - https://manpages.debian.org/testing/mlocate/mlocate.db.5.en.html
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh

        magic = int.from_bytes(self.fh.read(8), byteorder="big")
        if magic != c_mlocate.MAGIC:
            raise ValueError(f"Invalid mlocate file magic. Expected b'x00mlocate', got {magic}")

        self.header = c_mlocate.header_config(self.fh)

    def _parse_directory_entries(self) -> Iterator[str, c_mlocate.entry]:
        while (dbe_type := c_mlocate.DBE_TYPE(self.fh)) != c_mlocate.DBE_TYPE.END:
            entry = c_mlocate.entry(self.fh)
            dbe_type = "file" if dbe_type == c_mlocate.DBE_TYPE.FILE else "directory"

            yield dbe_type, entry

    def __iter__(self) -> Iterator[MLocateFile]:
        while True:
            try:
                directory_entry = c_mlocate.directory_entry(self.fh)
                parent = directory_entry.path.decode()

                for dbe_type, file_entry in self._parse_directory_entries():
                    file_path = file_entry.path.decode()

                    yield MLocate(
                        ts=from_unix(directory_entry.time_seconds),
                        ts_ns=directory_entry.time_nanoseconds,
                        parent=parent,
                        path=file_path,
                        dbe_type=dbe_type,
                    )
            except EOFError:  # noqa: PERF203
                return


class MLocatePlugin(BaseLocatePlugin):
    """Unix mlocate plugin."""

    __namespace__ = "mlocate"

    path = "/var/lib/mlocate/mlocate.db"

    def check_compatible(self) -> None:
        if not self.target.fs.path(self.path).exists():
            raise UnsupportedPluginError(f"No mlocate.db file found at {self.path}")

    @export(record=MLocateRecord)
    def locate(self) -> Iterator[MLocateRecord]:
        """Yield file and directory names from mlocate.db file.

        ``mlocate`` is a new implementation of GNU locate,
        but has been deprecated since Ubuntu 22.

        References:
            - https://manpages.debian.org/testing/mlocate/mlocate.db.5.en.html
        """
        mlocate_fh = self.target.fs.path(self.path).open()
        mlocate_file = MLocateFile(mlocate_fh)

        for item in mlocate_file:
            parent = self.target.fs.path(item.parent)
            yield MLocateRecord(
                ts=item.ts,
                ts_ns=item.ts_ns,
                parent=parent,
                path=parent.joinpath(item.path),
                type=item.dbe_type,
                source=self.path,
                _target=self.target,
            )
